from modelcommon.model import Store
from typing import Dict, Any
from pos_model import Order
from _NfceXmlPartBuilder import NfceXmlPartBuilder


class EmitBuilder(NfceXmlPartBuilder):
    def __init__(self, store_model):

        # type: (Store) -> None

        self.store_model = store_model

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        # Dados Emitente
        # CRT: 1 = simples nacional, 2 = simples (excesso de sublimite de recieta), 3 = regime normal
        inner_xml = "<emit>" \
                    "<CNPJ>%s</CNPJ>" \
                    "<xNome>%s</xNome>" \
                    "<enderEmit>" \
                    "<xLgr>%s</xLgr>" \
                    "<nro>%s</nro>" % \
                    (
                        self.store_model.cnpj,
                        self.store_model.name,
                        self.store_model.address_street,
                        self.store_model.address_number
                    )

        if self.store_model.address_complement:
            inner_xml += "<xCpl>%s</xCpl>" % self.store_model.address_complement

        inner_xml += "<xBairro>%s</xBairro>" \
               "<cMun>%s</cMun>" \
               "<xMun>%s</xMun>" \
               "<UF>%s</UF>" \
               "<CEP>%s</CEP>" \
               "</enderEmit>" \
               "<IE>%s</IE>" \
               "<CRT>%d</CRT>" \
               "</emit>" % \
                     (
                         self.store_model.neighborhood,
                         self.store_model.city_code,
                         self.store_model.city,
                         self.store_model.state,
                         self.store_model.postal_code,
                         self.store_model.state_registration,
                         self.store_model.crt
                     )

        return inner_xml



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.  All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file or at
# https://developers.google.com/open-source/licenses/bsd

"""A simple wrapper around enum types to expose utility functions.

Instances are created as properties with the same name as the enum they wrap
on proto classes.  For usage, see:
  reflection_test.py
"""

__author__ = 'rabsatt@google.com (Kevin Rabsatt)'


class EnumTypeWrapper(object):
  """A utility for finding the names of enum values."""

  DESCRIPTOR = None

  # This is a type alias, which mypy typing stubs can type as
  # a genericized parameter constrained to an int, allowing subclasses
  # to be typed with more constraint in .pyi stubs
  # Eg.
  # def MyGeneratedEnum(Message):
  #   ValueType = NewType('ValueType', int)
  #   def Name(self, number: MyGeneratedEnum.ValueType) -> str
  ValueType = int

  def __init__(self, enum_type):
    """Inits EnumTypeWrapper with an EnumDescriptor."""
    self._enum_type = enum_type
    self.DESCRIPTOR = enum_type  # pylint: disable=invalid-name

  def Name(self, number):  # pylint: disable=invalid-name
    """Returns a string containing the name of an enum value."""
    try:
      return self._enum_type.values_by_number[number].name
    except KeyError:
      pass  # fall out to break exception chaining

    if not isinstance(number, int):
      raise TypeError(
          'Enum value for {} must be an int, but got {} {!r}.'.format(
              self._enum_type.name, type(number), number))
    else:
      # repr here to handle the odd case when you pass in a boolean.
      raise ValueError('Enum {} has no name defined for value {!r}'.format(
          self._enum_type.name, number))

  def Value(self, name):  # pylint: disable=invalid-name
    """Returns the value corresponding to the given enum name."""
    try:
      return self._enum_type.values_by_name[name].number
    except KeyError:
      pass  # fall out to break exception chaining
    raise ValueError('Enum {} has no value defined for name {!r}'.format(
        self._enum_type.name, name))

  def keys(self):
    """Return a list of the string names in the enum.

    Returns:
      A list of strs, in the order they were defined in the .proto file.
    """

    return [value_descriptor.name
            for value_descriptor in self._enum_type.values]

  def values(self):
    """Return a list of the integer values in the enum.

    Returns:
      A list of ints, in the order they were defined in the .proto file.
    """

    return [value_descriptor.number
            for value_descriptor in self._enum_type.values]

  def items(self):
    """Return a list of the (name, value) pairs of the enum.

    Returns:
      A list of (str, int) pairs, in the order they were defined
      in the .proto file.
    """
    return [(value_descriptor.name, value_descriptor.number)
            for value_descriptor in self._enum_type.values]

  def __getattr__(self, name):
    """Returns the value corresponding to the given enum name."""
    try:
      return super(
          EnumTypeWrapper,
          self).__getattribute__('_enum_type').values_by_name[name].number
    except KeyError:
      pass  # fall out to break exception chaining
    raise AttributeError('Enum {} has no value defined for name {!r}'.format(
        self._enum_type.name, name))



from datetime import timedelta

from dateutil import tz

UTC = tz.UTC


def get_timezone(key):
    if not key:
        raise KeyError("Unknown time zone: %s" % key)

    try:
        rv = tz.gettz(key)
    except Exception:
        rv = None

    if rv is None or not isinstance(rv, (tz.tzutc, tz.tzfile)):
        raise KeyError("Unknown time zone: %s" % key)

    return rv


def get_timezone_file(f, key=None):
    return tz.tzfile(f)


def get_fixed_offset_zone(offset):
    return tz.tzoffset(None, timedelta(minutes=offset))


def is_ambiguous(dt):
    return tz.datetime_ambiguous(dt)


def is_imaginary(dt):
    return not tz.datetime_exists(dt)


enfold = tz.enfold


def get_fold(dt):
    return getattr(dt, "fold", 0)



from msgbus import MBEasyContext
from old_helper import BaseRepository
from persistence import Connection
from typing import (
    Dict,
    List,
    Tuple,
)


class FiscalParameterRepository(BaseRepository):

    def __init__(
        self,           # type: FiscalParameterRepository
        mb_context,     # type: MBEasyContext
    ):
        # type: (...) -> None
        super(FiscalParameterRepository, self).__init__(mb_context)

    def get_fiscal_parameters(
        self,           # type: FiscalParameterRepository
        loader_params,  # type: List[Dict[str, str]]
    ):
        # type: (...) -> Dict[Tuple[str, str], str]

        def inner_func(
            conn,   # type: Connection
        ):
            # type: (...) -> Dict[Tuple[int, str], str]

            query = "select ProductCode, ParamName, ParamValue from FiscalParameter"
            return {(str(x.get_entry(0)), x.get_entry(1)): x.get_entry(2) for x in conn.select(query)}

        fiscal_params = self.execute_with_connection(
            method=inner_func,
            service_name="FiscalPersistence",
        )

        if loader_params:
            for product in loader_params:
                product_code = product['PRODUCT_CODE']
                for param in product:
                    if param in ('PRODUCT_CODE', 'PRODUCT_NAME'):
                        continue
                    result = {(str(product_code), str(param)): str(product[param])}
                    fiscal_params.update(result)

        return fiscal_params



from datetime import (
    date,
    datetime,
)
from typing import (
    Dict,
    List,
    Optional,
)
from xml.etree import ElementTree

import iso8601
from order_api.dto import (
    Order,
    parse_order,
)
from ._ServiceChild import ServiceChild
from ._SimpleTable import SimpleTable
from ._TableMode import TableMode
from ._TableStatus import TableStatus


class TableServiceDto(object):
    def __init__(
        self,                       # type: TableServiceDto
        table,                      # type: SimpleTable
        service_id,                 # type: int
        mode,                       # type: TableMode
        started_at,                 # type: datetime
        business_period,            # type: date
        user_id,                    # type: int
        number_of_seats,            # type: Optional[int]
        custom_properties=None,     # type: Optional[Dict[str, str]]
        parent_table_id=None,       # type: Optional[str]
        source_table_id=None,       # type: Optional[str]
        service_orders=None,        # type: Optional[List[Order]]
        payment_order=None,         # type: Optional[Order]
        service_children=None,      # type: Optional[List[ServiceChild]]
    ):
        self.table = table
        self.service_id = service_id
        self.mode = mode
        self.status = table.status
        self.started_at = started_at
        self.business_period = business_period
        self.user_id = user_id
        self.number_of_seats = number_of_seats
        self.custom_properties = custom_properties or {}
        self.parent_table_id = parent_table_id
        self.source_table_id = source_table_id
        self.service_orders = service_orders or []
        self.payment_order = payment_order or []
        self.service_children = service_children or []

        self.updated_at = started_at

    def to_xml(self):
        table_attrib = self.table.get_table_xml_attrib()
        table_attrib.update({
            "sector": self.table.sector or "",
            "serviceId": str(self.service_id),
            "businessPeriod": self.business_period.strftime("%Y-%m-%d"),
            "userId": str(self.user_id),
            "startTS": self.started_at.strftime("%Y-%m-%d %H:%M:%S"),
            "lastUpdateTS": self.updated_at.strftime("%Y-%m-%d %H:%M:%S"),
            "mode": str(self.mode.value),
        })
        if self.number_of_seats is not None:
            table_attrib["serviceSeats"] = str(self.number_of_seats)

        table = ElementTree.Element(
            "Table",
            attrib=table_attrib,
        )

        if self.parent_table_id is not None:
            table.attrib["sliceParentTableId"] = str(self.parent_table_id)

        if self.source_table_id is not None:
            table.attrib["sliceSourceTableId"] = str(self.source_table_id)

        if len(self.service_orders) > 0:
            table.attrib["orders"] = ",".join(map(lambda x: str(x.order_id), self.service_orders))

            service_orders_xml = ElementTree.Element("ServiceOrders")
            orders_xml = ElementTree.Element("Orders")
            for service_order in self.service_orders:
                service_order_xml = ElementTree.Element(
                    "ServiceOrder",
                    attrib={
                        "orderId": str(service_order.order_id),
                        "state": service_order.state.name,
                        "tableServiceId": str(self.service_id),
                        "createdAt": service_order.created_at.strftime("%Y-%m-%dT%H:%M:%S.%f"),
                    }
                )
                service_orders_xml.append(service_order_xml)
                orders_xml.append(service_order.to_xml())
            table.append(service_orders_xml)
            table.append(orders_xml)

        if self.table.status == TableStatus.Totaled and self.payment_order:
            table.attrib["orders"] = str(self.payment_order.order_id)

            payment_orders_xml = ElementTree.Element("PaymentOrders")

            orders_xml = ElementTree.Element("Orders")
            payment_order_xml = ElementTree.Element(
                "PaymentOrder",
                attrib={
                    "orderId": str(self.payment_order.order_id),
                    "state": self.payment_order.state.name,
                    "createdAt": self.payment_order.created_at.strftime("%Y-%m-%dT%H:%M:%S.%f"),
                }
            )
            payment_orders_xml.append(payment_order_xml)
            orders_xml.append(self.payment_order.to_xml())

            table.append(payment_orders_xml)
            table.append(orders_xml)

        if self.service_children:
            service_children_xml = ElementTree.Element("ServiceChildren")
            for service_child in self.service_children:
                service_child_xml = ElementTree.Element(
                    "ServiceChild",
                    attrib={
                        "serviceId": str(service_child.service_id),
                        "tableId": service_child.table_id,
                        "status": str(service_child.status),
                        "tableName": service_child.table_name,
                    }
                )
                service_children_xml.append(service_child_xml)
            table.append(service_children_xml)

        if self.table.pos_id is not None:
            table.attrib["currentPOSId"] = str(self.table.pos_id)

        to_export_properties = {
            "MODE": "seats" if self.mode == TableMode.SEAT else "tabs",
        }
        for key, value in self.custom_properties.items():
            to_export_properties[key] = str(value)

        properties = ElementTree.Element(
            "Properties",
            attrib=to_export_properties,
        )
        table.append(properties)

        return table

    @staticmethod
    def from_xml(
        xml,        # type: ElementTree.Element
    ):
        # type: (...) -> TableServiceDto

        custom_properties = {}
        xml_properties = xml.find("Properties")
        for key, value in xml_properties.attrib.items():
            custom_properties[key] = value

        service_children = []
        for service_child_xml in xml.findall("ServiceChildren/ServiceChild"):
            service_children.append(ServiceChild(
                service_id=int(service_child_xml.attrib["serviceId"]),
                table_id=service_child_xml.attrib["tableId"],
                status=int(service_child_xml.attrib["status"]),
                table_name=service_child_xml.attrib["tableName"],
            ))

        service_orders = []
        service_order_ids = set()
        for order_xml in xml.findall("ServiceOrders/ServiceOrder"):
            service_order_ids.add(int(order_xml.attrib["orderId"]))

        payment_order = None
        payment_order_id = None
        for order_xml in xml.findall("PaymentOrders/PaymentOrder"):
            payment_order_id = int(order_xml.attrib["orderId"])

        for order_xml in xml.findall("Orders/Order"):
            order_id = int(order_xml.attrib["orderId"])
            if order_id in service_order_ids:
                service_orders.append(parse_order(order_xml))
            if order_id == payment_order_id:
                payment_order = parse_order(order_xml=order_xml)

        table = SimpleTable.from_xml(xml=xml)

        parent_table_id = xml.get("sliceParentTableId", None)
        source_table_id = xml.get("sliceSourceTableId", None)

        return TableServiceDto(
            table=table,
            service_id=int(xml.attrib["serviceId"]),
            mode=TableMode(int(xml.attrib["mode"])),
            started_at=iso8601.parse_date(xml.attrib["startTS"]),
            business_period=iso8601.parse_date(xml.attrib["businessPeriod"]).date(),
            user_id=int(xml.attrib["userId"]),
            number_of_seats=int(xml.attrib.get("serviceSeats")) if xml.get("serviceSeats") else None,
            custom_properties=custom_properties,
            service_children=service_children,
            service_orders=service_orders,
            payment_order=payment_order,
            parent_table_id=parent_table_id,
            source_table_id=source_table_id,
        )

    def get_tab_id(
        self,   # type: TableServiceDto
    ):
        # type: (...) -> str

        return "TAB-{}".format(self.table.table_id) if self.table.is_tab else None



# -*- coding: utf-8 -*-

import json

from actions import mb_context, logger
from bustoken import TK_CHAT_SEND_MESSAGE
from msgbus import FM_PARAM, TK_SYS_ACK, MBTimeout, MBException
from sysactions import action


@action
def send_chat_messages(_, message):
    message_data = json.dumps(dict(text=message))
    try:
        msg = mb_context.MB_EasySendMessage("ChatController", TK_CHAT_SEND_MESSAGE, FM_PARAM, message_data)
        if msg.token == TK_SYS_ACK:
            return True
    except (MBTimeout, MBException) as _:
        logger.exception("Error sending message to ChatController")
    
    return False



# Copyright 2009-2017 Wander Lairson Costa
# Copyright 2009-2020 PyUSB contributors
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

r"""PyUSB - Easy USB access in Python

This package exports the following modules and subpackages:

    core - the main USB implementation
    legacy - the compatibility layer with 0.x version
    backend - the support for backend implementations.
    control - USB standard control requests.
    libloader - helper module for backend library loading.

Since version 1.0, main PyUSB implementation lives in the 'usb.core'
module. New applications are encouraged to use it.
"""

import logging
import os

__author__ = 'Wander Lairson Costa'

# Use Semantic Versioning, http://semver.org/
try:
    from usb._version import version as __version__
except ImportError:
    __version__ = '0.0.0'

def _get_extended_version_info(version):
    import re
    m = re.match(r'(\d+)\.(\d+)\.(\d+)[.-]?(.*)', version)
    major, minor, patch, suffix = m.groups()
    return int(major), int(minor), int(patch), suffix

extended_version_info = _get_extended_version_info(__version__)
version_info = extended_version_info[:3]

__all__ = ['legacy', 'control', 'core', 'backend', 'util', 'libloader']

def _setup_log():
    from usb import _debug
    logger = logging.getLogger('usb')
    debug_level = os.getenv('PYUSB_DEBUG')

    if debug_level is not None:
        _debug.enable_tracing(True)
        filename = os.getenv('PYUSB_LOG_FILENAME')

        LEVELS = {'debug': logging.DEBUG,
                  'info': logging.INFO,
                  'warning': logging.WARNING,
                  'error': logging.ERROR,
                  'critical': logging.CRITICAL}

        level = LEVELS.get(debug_level, logging.CRITICAL + 10)
        logger.setLevel(level = level)

        try:
            handler = logging.FileHandler(filename)
        except:
            handler = logging.StreamHandler()

        fmt = logging.Formatter('%(asctime)s %(levelname)s:%(name)s:%(message)s')
        handler.setFormatter(fmt)
        logger.addHandler(handler)
    else:
        class NullHandler(logging.Handler):
            def emit(self, record):
                pass

        # We set the log level to avoid delegation to the
        # parent log handler (if there is one).
        # Thanks to Chris Clark to pointing this out.
        logger.setLevel(logging.CRITICAL + 10)

        logger.addHandler(NullHandler())


_setup_log()

# We import all 'legacy' module symbols to provide compatibility
# with applications that use 0.x versions.
from usb.legacy import *



import threading
from enum import Enum

from sysactions import get_model

local_thread_data = threading.local()


class ThreadLocalKey(Enum):
    model = "model"
    pos_id = "pos_id"
    user_level = "user_level"
    order_service = "order_service"
    pos_service = "pos_service"


def get_thread_var(varname):
    var = getattr(local_thread_data, varname, None)
    return var


def set_thread_var(varname, value):
    setattr(local_thread_data, varname, value)


def get_current_pos_id():
    if get_thread_var("pos_id"):
        return get_thread_var("pos_id")

    return None


def get_current_model():
    if get_thread_var("model"):
        return get_thread_var("model")

    model = get_model(get_current_pos_id())
    set_thread_var("model", model)
    return model



from xml.etree import cElementTree as eTree

from posot import OrderTakerException
from sysactions import (
    action,
    show_messagebox,
)

from actions.discount.clean_discounts import clean_order_payments
from actions.order.must_sync_ui import must_sync_ui
from api.order import get_order_service
from thread_local import get_current_model


@action
def do_back_from_total(pos_id, *args):
    model = get_current_model()
    order = model.find("CurrentOrder/Order")

    try:
        must_sync_ui(
            pos_id=pos_id,
            different_than="TOTALED",
        )
        clean_order_payments(
            pos_id=pos_id,
            order=order,
            check_manual_discounts=True
        )

        order_service = get_order_service(pos_id=pos_id)
        xml_order = order_service.reopen_order()

        return eTree.tostring(xml_order).decode("utf-8")

    except OrderTakerException as ex:
        show_messagebox(
            posid=pos_id,
            message="$ERROR_CODE_INFO|%d|%s" % (ex.getErrorCode(), ex.getErrorDescr()),
        )



# orm/state_changes.py
# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php

"""State tracking utilities used by :class:`_orm.Session`.

"""

from __future__ import annotations

import contextlib
from enum import Enum
from typing import Any
from typing import Callable
from typing import cast
from typing import Iterator
from typing import NoReturn
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

from .. import exc as sa_exc
from .. import util
from ..util.typing import Literal

_F = TypeVar("_F", bound=Callable[..., Any])


class _StateChangeState(Enum):
    pass


class _StateChangeStates(_StateChangeState):
    ANY = 1
    NO_CHANGE = 2
    CHANGE_IN_PROGRESS = 3


class _StateChange:
    """Supplies state assertion decorators.

    The current use case is for the :class:`_orm.SessionTransaction` class. The
    :class:`_StateChange` class itself is agnostic of the
    :class:`_orm.SessionTransaction` class so could in theory be generalized
    for other systems as well.

    """

    _next_state: _StateChangeState = _StateChangeStates.ANY
    _state: _StateChangeState = _StateChangeStates.NO_CHANGE
    _current_fn: Optional[Callable[..., Any]] = None

    def _raise_for_prerequisite_state(
        self, operation_name: str, state: _StateChangeState
    ) -> NoReturn:
        raise sa_exc.IllegalStateChangeError(
            f"Can't run operation '{operation_name}()' when Session "
            f"is in state {state!r}",
            code="isce",
        )

    @classmethod
    def declare_states(
        cls,
        prerequisite_states: Union[
            Literal[_StateChangeStates.ANY], Tuple[_StateChangeState, ...]
        ],
        moves_to: _StateChangeState,
    ) -> Callable[[_F], _F]:
        """Method decorator declaring valid states.

        :param prerequisite_states: sequence of acceptable prerequisite
         states.   Can be the single constant _State.ANY to indicate no
         prerequisite state

        :param moves_to: the expected state at the end of the method, assuming
         no exceptions raised.   Can be the constant _State.NO_CHANGE to
         indicate state should not change at the end of the method.

        """
        assert prerequisite_states, "no prequisite states sent"
        has_prerequisite_states = (
            prerequisite_states is not _StateChangeStates.ANY
        )

        prerequisite_state_collection = cast(
            "Tuple[_StateChangeState, ...]", prerequisite_states
        )
        expect_state_change = moves_to is not _StateChangeStates.NO_CHANGE

        @util.decorator
        def _go(fn: _F, self: Any, *arg: Any, **kw: Any) -> Any:
            current_state = self._state

            if (
                has_prerequisite_states
                and current_state not in prerequisite_state_collection
            ):
                self._raise_for_prerequisite_state(fn.__name__, current_state)

            next_state = self._next_state
            existing_fn = self._current_fn
            expect_state = moves_to if expect_state_change else current_state

            if (
                # destination states are restricted
                next_state is not _StateChangeStates.ANY
                # method seeks to change state
                and expect_state_change
                # destination state incorrect
                and next_state is not expect_state
            ):
                if existing_fn and next_state in (
                    _StateChangeStates.NO_CHANGE,
                    _StateChangeStates.CHANGE_IN_PROGRESS,
                ):
                    raise sa_exc.IllegalStateChangeError(
                        f"Method '{fn.__name__}()' can't be called here; "
                        f"method '{existing_fn.__name__}()' is already "
                        f"in progress and this would cause an unexpected "
                        f"state change to {moves_to!r}",
                        code="isce",
                    )
                else:
                    raise sa_exc.IllegalStateChangeError(
                        f"Cant run operation '{fn.__name__}()' here; "
                        f"will move to state {moves_to!r} where we are "
                        f"expecting {next_state!r}",
                        code="isce",
                    )

            self._current_fn = fn
            self._next_state = _StateChangeStates.CHANGE_IN_PROGRESS
            try:
                ret_value = fn(self, *arg, **kw)
            except:
                raise
            else:
                if self._state is expect_state:
                    return ret_value

                if self._state is current_state:
                    raise sa_exc.IllegalStateChangeError(
                        f"Method '{fn.__name__}()' failed to "
                        "change state "
                        f"to {moves_to!r} as expected",
                        code="isce",
                    )
                elif existing_fn:
                    raise sa_exc.IllegalStateChangeError(
                        f"While method '{existing_fn.__name__}()' was "
                        "running, "
                        f"method '{fn.__name__}()' caused an "
                        "unexpected "
                        f"state change to {self._state!r}",
                        code="isce",
                    )
                else:
                    raise sa_exc.IllegalStateChangeError(
                        f"Method '{fn.__name__}()' caused an unexpected "
                        f"state change to {self._state!r}",
                        code="isce",
                    )

            finally:
                self._next_state = next_state
                self._current_fn = existing_fn

        return _go

    @contextlib.contextmanager
    def _expect_state(self, expected: _StateChangeState) -> Iterator[Any]:
        """called within a method that changes states.

        method must also use the ``@declare_states()`` decorator.

        """
        assert self._next_state is _StateChangeStates.CHANGE_IN_PROGRESS, (
            "Unexpected call to _expect_state outside of "
            "state-changing method"
        )

        self._next_state = expected
        try:
            yield
        except:
            raise
        else:
            if self._state is not expected:
                raise sa_exc.IllegalStateChangeError(
                    f"Unexpected state change to {self._state!r}", code="isce"
                )
        finally:
            self._next_state = _StateChangeStates.CHANGE_IN_PROGRESS



from collections.abc import Callable, Hashable, Iterator
from functools import partial
from operator import methodcaller
import sys
from typing import (
    Any,
    Generic,
    Protocol,
    TypeVar,
    overload,
)

if sys.version_info >= (3, 10):
    from typing import Concatenate, ParamSpec
else:
    from typing_extensions import Concatenate, ParamSpec

_P = ParamSpec('_P')
_R = TypeVar('_R')
_T = TypeVar('_T')
_R1 = TypeVar('_R1')
_R2 = TypeVar('_R2')
_V = TypeVar('_V')
_S = TypeVar('_S')
_R_co = TypeVar('_R_co', covariant=True)

class _OnceCallable(Protocol[_P, _R]):
    saved_result: _R
    reset: Callable[[], None]
    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _R: ...

class _ProxyMethodCacheWrapper(Protocol[_R_co]):
    cache_clear: Callable[[], None]
    def __call__(self, *args: Hashable, **kwargs: Hashable) -> _R_co: ...

class _MethodCacheWrapper(Protocol[_R_co]):
    def cache_clear(self) -> None: ...
    def __call__(self, *args: Hashable, **kwargs: Hashable) -> _R_co: ...

# `compose()` overloads below will cover most use cases.

@overload
def compose(
    __func1: Callable[[_R], _T],
    __func2: Callable[_P, _R],
    /,
) -> Callable[_P, _T]: ...
@overload
def compose(
    __func1: Callable[[_R], _T],
    __func2: Callable[[_R1], _R],
    __func3: Callable[_P, _R1],
    /,
) -> Callable[_P, _T]: ...
@overload
def compose(
    __func1: Callable[[_R], _T],
    __func2: Callable[[_R2], _R],
    __func3: Callable[[_R1], _R2],
    __func4: Callable[_P, _R1],
    /,
) -> Callable[_P, _T]: ...
def once(func: Callable[_P, _R]) -> _OnceCallable[_P, _R]: ...
def method_cache(
    method: Callable[..., _R],
    cache_wrapper: Callable[[Callable[..., _R]], _MethodCacheWrapper[_R]] = ...,
) -> _MethodCacheWrapper[_R] | _ProxyMethodCacheWrapper[_R]: ...
def apply(
    transform: Callable[[_R], _T]
) -> Callable[[Callable[_P, _R]], Callable[_P, _T]]: ...
def result_invoke(
    action: Callable[[_R], Any]
) -> Callable[[Callable[_P, _R]], Callable[_P, _R]]: ...
def invoke(
    f: Callable[_P, _R], /, *args: _P.args, **kwargs: _P.kwargs
) -> Callable[_P, _R]: ...

class Throttler(Generic[_R]):
    last_called: float
    func: Callable[..., _R]
    max_rate: float
    def __init__(
        self, func: Callable[..., _R] | Throttler[_R], max_rate: float = ...
    ) -> None: ...
    def reset(self) -> None: ...
    def __call__(self, *args: Any, **kwargs: Any) -> _R: ...
    def __get__(self, obj: Any, owner: type[Any] | None = ...) -> Callable[..., _R]: ...

def first_invoke(
    func1: Callable[..., Any], func2: Callable[_P, _R]
) -> Callable[_P, _R]: ...

method_caller: Callable[..., methodcaller]

def retry_call(
    func: Callable[..., _R],
    cleanup: Callable[..., None] = ...,
    retries: int | float = ...,
    trap: type[BaseException] | tuple[type[BaseException], ...] = ...,
) -> _R: ...
def retry(
    cleanup: Callable[..., None] = ...,
    retries: int | float = ...,
    trap: type[BaseException] | tuple[type[BaseException], ...] = ...,
) -> Callable[[Callable[..., _R]], Callable[..., _R]]: ...
def print_yielded(func: Callable[_P, Iterator[Any]]) -> Callable[_P, None]: ...
def pass_none(
    func: Callable[Concatenate[_T, _P], _R]
) -> Callable[Concatenate[_T, _P], _R]: ...
def assign_params(
    func: Callable[..., _R], namespace: dict[str, Any]
) -> partial[_R]: ...
def save_method_args(
    method: Callable[Concatenate[_S, _P], _R]
) -> Callable[Concatenate[_S, _P], _R]: ...
def except_(
    *exceptions: type[BaseException], replace: Any = ..., use: Any = ...
) -> Callable[[Callable[_P, Any]], Callable[_P, Any]]: ...
def identity(x: _T) -> _T: ...
def bypass_when(
    check: _V, *, _op: Callable[[_V], Any] = ...
) -> Callable[[Callable[[_T], _R]], Callable[[_T], _T | _R]]: ...
def bypass_unless(
    check: Any,
) -> Callable[[Callable[[_T], _R]], Callable[[_T], _T | _R]]: ...



from datetime import datetime

from ._OrderState import OrderState


class StateHistory(object):
    def __init__(
        self,           # type: StateHistory
        state,          # type: OrderState
        created_at,     # type: datetime
    ):
        self.state = state
        self.created_at = created_at



import os
import re
import urllib.parse
import urllib.request

import pytest

from setuptools import Distribution
from setuptools.dist import check_package_data, check_specifier

from .test_easy_install import make_nspkg_sdist
from .test_find_packages import ensure_files
from .textwrap import DALS

from distutils.errors import DistutilsSetupError


def test_dist_fetch_build_egg(tmpdir):
    """
    Check multiple calls to `Distribution.fetch_build_egg` work as expected.
    """
    index = tmpdir.mkdir('index')
    index_url = urllib.parse.urljoin('file://', urllib.request.pathname2url(str(index)))

    def sdist_with_index(distname, version):
        dist_dir = index.mkdir(distname)
        dist_sdist = '%s-%s.tar.gz' % (distname, version)
        make_nspkg_sdist(str(dist_dir.join(dist_sdist)), distname, version)
        with dist_dir.join('index.html').open('w') as fp:
            fp.write(
                DALS(
                    """
                <!DOCTYPE html><html><body>
                <a href="{dist_sdist}" rel="internal">{dist_sdist}</a><br/>
                </body></html>
                """
                ).format(dist_sdist=dist_sdist)
            )

    sdist_with_index('barbazquux', '3.2.0')
    sdist_with_index('barbazquux-runner', '2.11.1')
    with tmpdir.join('setup.cfg').open('w') as fp:
        fp.write(
            DALS(
                """
            [easy_install]
            index_url = {index_url}
            """
            ).format(index_url=index_url)
        )
    reqs = """
    barbazquux-runner
    barbazquux
    """.split()
    with tmpdir.as_cwd():
        dist = Distribution()
        dist.parse_config_files()
        resolved_dists = [dist.fetch_build_egg(r) for r in reqs]
    assert [dist.key for dist in resolved_dists if dist] == reqs


EXAMPLE_BASE_INFO = dict(
    name="package",
    version="0.0.1",
    author="Foo Bar",
    author_email="foo@bar.net",
    long_description="Long\ndescription",
    description="Short description",
    keywords=["one", "two"],
)


def test_provides_extras_deterministic_order():
    attrs = dict(extras_require=dict(a=['foo'], b=['bar']))
    dist = Distribution(attrs)
    assert list(dist.metadata.provides_extras) == ['a', 'b']
    attrs['extras_require'] = dict(reversed(attrs['extras_require'].items()))
    dist = Distribution(attrs)
    assert list(dist.metadata.provides_extras) == ['b', 'a']


CHECK_PACKAGE_DATA_TESTS = (
    # Valid.
    (
        {
            '': ['*.txt', '*.rst'],
            'hello': ['*.msg'],
        },
        None,
    ),
    # Not a dictionary.
    (
        (
            ('', ['*.txt', '*.rst']),
            ('hello', ['*.msg']),
        ),
        (
            "'package_data' must be a dictionary mapping package"
            " names to lists of string wildcard patterns"
        ),
    ),
    # Invalid key type.
    (
        {
            400: ['*.txt', '*.rst'],
        },
        ("keys of 'package_data' dict must be strings (got 400)"),
    ),
    # Invalid value type.
    (
        {
            'hello': '*.msg',
        },
        (
            "\"values of 'package_data' dict\" must be of type <tuple[str, ...] | list[str]>"
            " (got '*.msg')"
        ),
    ),
    # Invalid value type (generators are single use)
    (
        {
            'hello': (x for x in "generator"),
        },
        (
            "\"values of 'package_data' dict\" must be of type <tuple[str, ...] | list[str]>"
            " (got <generator object"
        ),
    ),
)


@pytest.mark.parametrize('package_data, expected_message', CHECK_PACKAGE_DATA_TESTS)
def test_check_package_data(package_data, expected_message):
    if expected_message is None:
        assert check_package_data(None, 'package_data', package_data) is None
    else:
        with pytest.raises(DistutilsSetupError, match=re.escape(expected_message)):
            check_package_data(None, 'package_data', package_data)


def test_check_specifier():
    # valid specifier value
    attrs = {'name': 'foo', 'python_requires': '>=3.0, !=3.1'}
    dist = Distribution(attrs)
    check_specifier(dist, attrs, attrs['python_requires'])

    # invalid specifier value
    attrs = {'name': 'foo', 'python_requires': ['>=3.0', '!=3.1']}
    with pytest.raises(DistutilsSetupError):
        dist = Distribution(attrs)


def test_metadata_name():
    with pytest.raises(DistutilsSetupError, match='missing.*name'):
        Distribution()._validate_metadata()


@pytest.mark.parametrize(
    "dist_name, py_module",
    [
        ("my.pkg", "my_pkg"),
        ("my-pkg", "my_pkg"),
        ("my_pkg", "my_pkg"),
        ("pkg", "pkg"),
    ],
)
def test_dist_default_py_modules(tmp_path, dist_name, py_module):
    (tmp_path / f"{py_module}.py").touch()

    (tmp_path / "setup.py").touch()
    (tmp_path / "noxfile.py").touch()
    # ^-- make sure common tool files are ignored

    attrs = {**EXAMPLE_BASE_INFO, "name": dist_name, "src_root": str(tmp_path)}
    # Find `py_modules` corresponding to dist_name if not given
    dist = Distribution(attrs)
    dist.set_defaults()
    assert dist.py_modules == [py_module]
    # When `py_modules` is given, don't do anything
    dist = Distribution({**attrs, "py_modules": ["explicity_py_module"]})
    dist.set_defaults()
    assert dist.py_modules == ["explicity_py_module"]
    # When `packages` is given, don't do anything
    dist = Distribution({**attrs, "packages": ["explicity_package"]})
    dist.set_defaults()
    assert not dist.py_modules


@pytest.mark.parametrize(
    "dist_name, package_dir, package_files, packages",
    [
        ("my.pkg", None, ["my_pkg/__init__.py", "my_pkg/mod.py"], ["my_pkg"]),
        ("my-pkg", None, ["my_pkg/__init__.py", "my_pkg/mod.py"], ["my_pkg"]),
        ("my_pkg", None, ["my_pkg/__init__.py", "my_pkg/mod.py"], ["my_pkg"]),
        ("my.pkg", None, ["my/pkg/__init__.py"], ["my", "my.pkg"]),
        (
            "my_pkg",
            None,
            ["src/my_pkg/__init__.py", "src/my_pkg2/__init__.py"],
            ["my_pkg", "my_pkg2"],
        ),
        (
            "my_pkg",
            {"pkg": "lib", "pkg2": "lib2"},
            ["lib/__init__.py", "lib/nested/__init__.pyt", "lib2/__init__.py"],
            ["pkg", "pkg.nested", "pkg2"],
        ),
    ],
)
def test_dist_default_packages(
    tmp_path, dist_name, package_dir, package_files, packages
):
    ensure_files(tmp_path, package_files)

    (tmp_path / "setup.py").touch()
    (tmp_path / "noxfile.py").touch()
    # ^-- should not be included by default

    attrs = {
        **EXAMPLE_BASE_INFO,
        "name": dist_name,
        "src_root": str(tmp_path),
        "package_dir": package_dir,
    }
    # Find `packages` either corresponding to dist_name or inside src
    dist = Distribution(attrs)
    dist.set_defaults()
    assert not dist.py_modules
    assert not dist.py_modules
    assert set(dist.packages) == set(packages)
    # When `py_modules` is given, don't do anything
    dist = Distribution({**attrs, "py_modules": ["explicit_py_module"]})
    dist.set_defaults()
    assert not dist.packages
    assert set(dist.py_modules) == {"explicit_py_module"}
    # When `packages` is given, don't do anything
    dist = Distribution({**attrs, "packages": ["explicit_package"]})
    dist.set_defaults()
    assert not dist.py_modules
    assert set(dist.packages) == {"explicit_package"}


@pytest.mark.parametrize(
    "dist_name, package_dir, package_files",
    [
        ("my.pkg.nested", None, ["my/pkg/nested/__init__.py"]),
        ("my.pkg", None, ["my/pkg/__init__.py", "my/pkg/file.py"]),
        ("my_pkg", None, ["my_pkg.py"]),
        ("my_pkg", None, ["my_pkg/__init__.py", "my_pkg/nested/__init__.py"]),
        ("my_pkg", None, ["src/my_pkg/__init__.py", "src/my_pkg/nested/__init__.py"]),
        (
            "my_pkg",
            {"my_pkg": "lib", "my_pkg.lib2": "lib2"},
            ["lib/__init__.py", "lib/nested/__init__.pyt", "lib2/__init__.py"],
        ),
        # Should not try to guess a name from multiple py_modules/packages
        ("UNKNOWN", None, ["src/mod1.py", "src/mod2.py"]),
        ("UNKNOWN", None, ["src/pkg1/__ini__.py", "src/pkg2/__init__.py"]),
    ],
)
def test_dist_default_name(tmp_path, dist_name, package_dir, package_files):
    """Make sure dist.name is discovered from packages/py_modules"""
    ensure_files(tmp_path, package_files)
    attrs = {
        **EXAMPLE_BASE_INFO,
        "src_root": "/".join(os.path.split(tmp_path)),  # POSIX-style
        "package_dir": package_dir,
    }
    del attrs["name"]

    dist = Distribution(attrs)
    dist.set_defaults()
    assert dist.py_modules or dist.packages
    assert dist.get_name() == dist_name



import logging

import sysactions

from actions.eft import confirm_finish_eft_transactions

logger = logging.getLogger("EftActions")


@sysactions.action
def confirm_cancel_eft_mobile_transactions(
    pos_id,             # type: str
    data,               # type: str
    result_code=None,   # type: str
    message=None        # type: str
):
    # type: (...) -> None

    try:
        if not data or result_code != "0":
            error = "Error confirm cancel eft mobile transactions: {} | {} | {}".format(result_code, message, data)
            logger.error(error)
            confirm_finish_eft_transactions(
                pos_id=pos_id,
                status="0",
            )
            return

        logger.debug("Success cancel transactions mobile: {0}".format(data))
        confirm_finish_eft_transactions(
            pos_id=pos_id,
            status="1",
            data=data,
        )
    except Exception as _:
        logger.exception("Error cancel transactions mobile: {0}".format(data))
        confirm_finish_eft_transactions(
            pos_id=pos_id,
            status="0",
        )



from api.order import get_order_service
from application.domain import OrderManager, SaleLine  # noqa


class MsgBusOrderManager(OrderManager):

    def recall_order(self, pos_id, order_id):
        order_service = get_order_service(pos_id=pos_id)
        order_service.recall_order(order_id=order_id)

    def store_order(self, pos_id):
        order_service = get_order_service(pos_id=pos_id)
        order_service.store_order()

    def get_current_order_state(self, pos_id):
        order_service = get_order_service(pos_id=pos_id)

        order_picture = order_service.get_order_picture()
        if order_picture is not None:
            return order_picture.get("state")
        return None



from sysactions import (
    show_messagebox,
    StopAction,
)


def show_nak_message(
    pos_id,         # type: str
    update_key,     # type: str
    message,        # type: str
):
    # type: (...) -> None

    response_message = message.upper()
    if '$' not in response_message:
        response_message = update_key + response_message

    show_messagebox(
        posid=pos_id,
        message=response_message,
        buttons="OK"
    )

    raise StopAction()



from typing import Optional

from actions.config import get_pre_account_allow_printing
from actions.printer import nreports_preview_and_print
from api.nreports import get_nreports_service
from sysactions import (
    action,
    get_model,
    get_current_order,
)


@action
def qsr_totaled_total_report(
    pos_id,                 # type: str
    is_mobile=False,        # type: Optional[bool]
):
    # type: (...) -> None

    model = get_model(posid=pos_id)
    order = get_current_order(model=model)
    is_mobile = is_mobile.lower() == "true" if isinstance(is_mobile, str) else is_mobile

    print_data = get_nreports_service().pre_account_quick_service(
        pos_id=pos_id,
        order_id=order.get("orderId"),
        is_mobile=is_mobile,
    )
    nreports_preview_and_print.print_preview_report(
        pos_id=pos_id,
        report_data=print_data,
        preview=True,
        order=order,
        allow_printing=get_pre_account_allow_printing(),
    )



import re
import cPickle as Pickle

from datetime import datetime

from helper import remove_accents
from typing import (
    List,
    Optional,
    Dict,
)
from collections import OrderedDict

from ._CustomStorage import CustomStorage
from ._TagEvent import TagEvent
from ._TagEventType import TagEventType


class Item(CustomStorage):

    """ class Item(CustomStorage)
    Represents an item of a production order.
    Instances of this class can be handled and/or modified by any production-box.
    Use the methods write_data, read_data and delete_data in order to modify instances
    in a safe way (avoiding name conflicts).
    """
    # Note - NEVER use mutable objects as the default value here (they are shared by all instances)
    # Mutable objects MUST be created in the constructor

    line_number = 0             #: Number of the "sale line" that represents this item
    item_id = ""                #: The "context" of this item (n levels)
    part_code = ""              #: The product code (one level)
    item_code = ""              #: "item_id.part_code" (n+1 levels)
    description = ""            #: The product name
    qty = 0                     #: Item quantity
    qty_added = 0               #: Total quantity added of this item
    qty_voided = 0              #: Total quantity voided of this item
    qty_modified = 0            #: Quantity modified (for components only)
    default_qty = 0             #: Default quantity of this item (only for sub-items)
    min_qty = 0                 #: Minimum quantity of this item (only for sub-items)
    max_qty = 0                 #: Maximum quantity of this item (only for sub-items)
    last_order_qty = 0          #: Last quantity of this item
    multiplied_qty = 0          #: Real (multiplied) item quantity (considers parent quantity)
    item_type = ""              #: Indicates if this item is an OPTION, etc
    level = 0                   #: Item level (0, 1, 2, ...)
    pos_level = 0               #: Item level from the complete order in the POS
    modifier_label = None       #: Optional modifier label
    only = ""                   #: Only modifier
    json_tags = ""              #: product tags as json text
    product_priority = 100      #: sorting items
    voided = False              #: If this item, or a parent item, was voided
    last_updated = ''           #: last time this item was changed in the POS
    main_item_description = ''  #: name of level 0 item
    seat = ''                   #: Item seat
    manual_course = ''          #: Item new production course
    items = ()                  #: List of child items
    comments = ()               #: item comments
    properties = ()             #: List of custom properties from the Order Picture
    prep_time = None
    send_moment = None
    tags = {}
    tag_history = ()
    cost = None
    count = 0
    total_count = 0
    measure_unit = ""
    jit_lines = ""
    fire = False
    high_priority = False
    show_removed = False
    original_level = None               # type: Optional[int]
    original_qty = None                 # type: Optional[float]
    sequence = 1000
    fraction = 1
    fraction_parent = None
    grouped_comment = None
    fractionated_by = None
    created_at = None
    custom_params = {}

    def __init__(self, **kargs):
        CustomStorage.__init__(self)
        self.items = list()
        self.comments = dict()
        self.properties = dict()
        self.tags = {}              # type: OrderedDict[str, str]
        self.tag_history = []       # type: List[TagEvent]
        self.joined_items = []      # type: List[Item]
        self.prep_time = None       # type: Optional[int]
        self.send_moment = None     # type: Optional[datetime]
        self.cost = None          # type: Optional[int]
        self.displayed_jit = []     # type: List[str]
        self.jit_lines = ""         # type: str
        self.block_tags = False     # type: bool
        self.show_removed = False   # type: bool
        self.custom_params = {}     # type: Dict

        for key, val in kargs.iteritems():
            setattr(self, key, val)

        ignore_keys = ["fiscalParameters", "sessionId", "plainPrice"]
        self.properties = {key: self.properties[key] for key in self.properties if key not in ignore_keys}

        if not self.item_code:
            self.item_code = "%s.%s" % (self.item_id, self.part_code)

    def __str__(self):
        item_description = " " * int(self.level) + str(self.qty) + " " + self.description.encode("utf-8")
        if self.has_tags():
            item_description += " ["
            for tag in self.tags:
                item_description += tag.encode("utf-8") + ", "
            item_description = item_description[:-2]
            item_description += "]"

        item_description += " [" + self.get_line_id() + "]"
        item_description += " {}".format(self.item_type)
        if self.prep_time is not None:
            item_description += " PrepTime: " + str(self.prep_time)
        if self.send_moment is not None:
            item_description += " SendMoment: " + str(self.send_moment.isoformat())
        if self.cost is not None:
            item_description += " Cost: " + str(self.cost)
        if len(self.joined_items) > 0:
            item_description += "\nJoined Items:"
            for joined_item in self.joined_items:
                item_description += "\n" + str(joined_item)
        for son in self.items:
            item_description += "\n" + str(son)

        return item_description

    def __repr__(self):
        return "(%s) %s" % (object.__repr__(self), str(self))

    def get_tags(self):
        return self.tags.keys()

    def get_sons_tags(self):
        tags = []

        for son in self.items:
            tags.extend(son.get_tags())
            tags.extend(son.get_sons_tags())

        return list(set(tags))

    def has_tags(self):
        return len(self.tags) > 0

    def has_tag(self, tag):
        return tag in self.tags

    def has_some_tag(self, tags):
        for tag in tags:
            if tag in self.tags:
                break
        else:
            return False

        return True

    def does_not_have_tag(self, tag):
        return tag not in self.tags

    def add_view(self, kds_view):
        if 'kds_view' in self.tags.keys() and self.tags['kds_view'] == kds_view:
            return False

        self.tags['kds_view'] = kds_view
        self.tag_history.append(TagEvent(kds_view, TagEventType.added, datetime.now()))
        return True

    def add_tag(self, tag, line_jit_list=None):
        if not line_jit_list:
            self._add_single_tag(tag)
            return

        for jit in line_jit_list:
            jit_tag = tag + "-" + jit.lower()
            self._add_single_tag(jit_tag)

    def _add_single_tag(self, tag):
        if tag not in self.tags:
            self.tags[tag] = tag
            self.tag_history.append(TagEvent(tag, TagEventType.added, datetime.now()))

    def remove_tag(self, tag, line_jit_list=None):
        if not line_jit_list:
            self._remove_single_tag(tag)
            return

        for jit in line_jit_list:
            jit_tag = tag + "-" + jit.lower()
            self._remove_single_tag(jit_tag)

    def _remove_single_tag(self, tag):
        if tag in self.tags:
            del self.tags[tag]
            self.tag_history.append(TagEvent(tag, TagEventType.removed, datetime.now()))

    def toggle_tag(self, tag, line_jit_list=None):
        if not line_jit_list:
            self._toggle_single_tag(tag)
            return [tag]

        jit_tags = []
        for jit in line_jit_list:
            jit_tag = tag + "-" + jit.lower()
            self._toggle_single_tag(jit_tag)
            jit_tags.append(jit_tag)

        return jit_tags

    def _toggle_single_tag(self, tag):
        if tag in self.tags:
            self._remove_single_tag(tag)
        else:
            self._add_single_tag(tag)

    def is_product(self):
        return self.item_type in ("PRODUCT", "CANADD")

    def is_ingredient(self):
        return self.item_type == "INGREDIENT"

    def is_combo(self):
        return self.item_type == "COMBO"

    def is_option(self):
        return self.item_type == "OPTION"

    def has_some_product_child(self):
        for son in self.items:
            if son.is_product() or son.has_some_product_child():
                return True

        return False

    def is_prepared(self):
        return self.get_last_prepared_time() is not None

    def get_last_prepared_time(self):
        if self.is_product():
            part_code_jit = self.jit_lines
            last_jit_sequence = re.split('[><$]', part_code_jit)[-1]

            necessary_jit_lines = re.findall('(\\w+)', last_jit_sequence)
            prepared_jit_tags = self.check_all_tag_in_item(necessary_jit_lines, ["done", "served", "printed"])
            prepared_generic_tags = self.check_all_tag_in_item(None, ["served", "hold"])
            tag_list = [self.get_last_tag_time(prepared_jit_tags), self.get_last_tag_time(prepared_generic_tags)]
            filtered_tag_list = [x for x in tag_list if x is not None]

            return max(filtered_tag_list) if len(filtered_tag_list) > 0 else None

        son_times = []
        for son in self.items:
            if son.qty == 0:
                continue

            son_last_prepared_tag = son.get_last_prepared_time()
            if son_last_prepared_tag is None:
                return None

            son_times.append(son_last_prepared_tag)

        return max(son_times) if son_times else None

    def all_has_at_least_one_tag(self, tags):
        return self.all_has_at_least_one_jit_tag(None, tags)

    def at_least_one_has_at_least_one_tag(self, allowed_tags):
        if self.ignore_removed_item() or not allowed_tags:
            return True

        if self.is_product():
            if not self.tags:
                return True

            has_item = False
            for tag in self.tags:
                if tag in allowed_tags:
                    has_item = True

            return has_item

        for son in self.items:
            if son.at_least_one_has_at_least_one_tag(allowed_tags):
                return True

        return False

    def check_all_tag_in_item(self, necessary_jit_lines, tag_types):
        if necessary_jit_lines:
            tag_types = ["{}-{}".format(x, y.lower()) for x in tag_types for y in necessary_jit_lines]

        return next((x for x in tag_types if x in self.tags), None)

    def all_has_at_least_one_jit_tag(self, necessary_jit_lines, tag_types):
        if self.ignore_removed_item():
            return True

        if self.is_product():
            if self.check_all_tag_in_item(necessary_jit_lines, tag_types) is not None:
                return True

            return False

        for son in self.items:
            if not son.all_has_at_least_one_jit_tag(necessary_jit_lines, tag_types):
                return False

        return True

    def get_last_tag_time(self, tag):
        # type: (str) -> Optional[datetime]

        if self.is_product():
            for event in reversed(self.tag_history):
                if event.tag == tag and event.action == TagEventType.added:
                    return event.date
            return None
        else:
            last_time = None
            for son in self.items:
                son_last_time = son.get_last_tag_time(tag)
                if son_last_time is not None and (last_time is None or son_last_time > last_time):
                    last_time = son_last_time
            return last_time

    def get_last_tag_list_time(self, tag_list):
        time_list = [self.get_last_tag_time(x) for x in tag_list]
        filtered_tag_list = [x for x in time_list if x is not None]

        return max(filtered_tag_list) if len(filtered_tag_list) > 0 else None

    def clone(self):
        return Pickle.loads(Pickle.dumps(self, -1))

    def get_line_id(self):
        part_code = str(self.part_code)
        if hasattr(self, "original_part_code"):
            part_code = str(self.original_part_code)
        level = str(int(self.level))
        if self.original_level is not None:
            level = str(int(self.original_level))

        return str(self.line_number) + "-" + self.item_id + "-" + level + "-" + part_code

    def ignore_removed_item(self):
        return self.qty == 0 and not self.show_removed

    def handle_remove_father(
        self,
        father,                     # type: Item
        inherit_comments=False,     # type: Optional[bool]
    ):
        # type: (...) -> None

        if father is None:
            return

        if inherit_comments and father.comments:
            self.repeat_comment_from_father(father=father)

        if self.original_qty is None:
            self.original_qty = self.qty

        self.seat = father.seat

        self.adjust_level()

        father_level = father.level if father.original_level is None else father.original_level
        if father_level == 0:
            self.qty *= father.qty
            return

        if father.original_qty is not None and father.qty != father.original_qty and father.original_qty != 0:
            self.qty *= father.qty / father.original_qty

    def repeat_comment_from_father(
        self,
        father,     # type: Item
    ):
        # type: (...) -> None

        option_comment = "(C) "
        for key in father.comments:
            comment = remove_accents(father.comments.get(key).text)
            comment_prefix = "" if comment.startswith(option_comment) else option_comment
            father.comments[key].text = "{}{}".format(comment_prefix, comment)
        self.comments.update(father.comments)

    def adjust_level(
        self,   # type: Item
    ):
        # type: (...) -> None

        if self.original_level is None:
            self.original_level = self.level

        self.level -= 1

        for son in self.items:
            son.adjust_level()



import sys
from abc import ABCMeta
from logging import Logger

from production.box.statistic import StatisticBox
from production.manager import PublishScheduler
from production.model import (
    StatisticClassNotFoundException,
    MissingBoxParameterException,
)
from production.treebuilder import BoxConfiguration
from typing import Dict

from ._BoxBuilder import BoxBuilder


class StatisticBoxBuilder(BoxBuilder):
    __metaclass__ = ABCMeta

    def __init__(
        self,  # type: StatisticBoxBuilder
        publish_scheduler,  # type: PublishScheduler
        loggers,  # type: Dict[str, Logger]
    ):
        # type: (...) -> None

        super(StatisticBoxBuilder, self).__init__(loggers)
        self.publish_scheduler = publish_scheduler

    def build(self, config):
        # type: (BoxConfiguration) -> StatisticBox

        def get_config(config_name, default=None):
            return self.get_extra(config, config_name, default)

        statistic_class_name = self.get_extra_config(config, "StatisticClass")
        statistic_type = getattr(sys.modules["production.box.statistic"], statistic_class_name, None)
        if statistic_type is None:
            raise StatisticClassNotFoundException("StatisticClass ({}) does not exist".format(statistic_class_name))

        return statistic_type(
            config.name,
            config.sons,
            self.publish_scheduler,
            self.get_logger(config),
            get_config
        )

    def get_extra_config(self, config, config_name, default=None):
        config_value = self.get_extra(config, config_name, default)
        if config_value is None:
            raise MissingBoxParameterException("({}) does not exist".format(config_name))

        return config_value



# iso11649.py - functions for performing the ISO 11649 checksum validation
#               for structured creditor reference numbers
#
# Copyright (C) 2018 Esben Toke Christensen
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""ISO 11649 (Structured Creditor Reference).

The ISO 11649 structured creditor number consists of 'RF' followed by two
check digits and up to 21 digits. The number may contain letters.

The reference number is validated by moving RF and the check digits to the
end of the number, and checking that the ISO 7064 Mod 97, 10 checksum of this
string is 1.

More information:

* https://en.wikipedia.org/wiki/Creditor_Reference

>>> validate('RF18 5390 0754 7034')
'RF18539007547034'
>>> validate('RF18 5390 0754 70Y')
'RF185390075470Y'
>>> is_valid('RF18 5390 0754 7034')
True
>>> validate('RF17 5390 0754 7034')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('RF18539007547034')
'RF18 5390 0754 7034'
"""

from stdnum.exceptions import *
from stdnum.iso7064 import mod_97_10
from stdnum.util import clean


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any invalid separators and removes surrounding whitespace."""
    return clean(number, ' -.,/:').upper().strip()


def validate(number):
    """Check if the number provided is a valid ISO 11649 structured creditor
    reference number."""
    number = compact(number)
    if len(number) < 5 or len(number) > 25:
        raise InvalidLength()
    if not number.startswith('RF'):
        raise InvalidFormat()
    mod_97_10.validate(number[4:] + number[:4])
    return number


def is_valid(number):
    """Check if the number provided is a valid ISO 11649 structured creditor
    number. This checks the length, formatting and check digits."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Format the number provided for output.

    Blocks of 4 characters, the last block can be less than 4 characters. See
    https://www.paymentstandards.ch/dam/downloads/ig-qr-bill-en.pdf chapter
    3.6.2.
    """
    number = compact(number)
    return ' '.join(number[i:i + 4] for i in range(0, len(number), 4))



import sys

import autocommand

from jaraco.text import Stripper


def strip_prefix():
    r"""
    Strip any common prefix from stdin.

    >>> import io, pytest
    >>> getfixture('monkeypatch').setattr('sys.stdin', io.StringIO('abcdef\nabc123'))
    >>> strip_prefix()
    def
    123
    """
    sys.stdout.writelines(Stripper.strip_prefix(sys.stdin).lines)


autocommand.autocommand(__name__)(strip_prefix)



from typing import List


class PartWithDependencies(object):
    def __init__(
        self,           # type: PartWithDependencies
        part_code,      # type: int
        name,           # type: str
        dependencies,   # type: List[int]
    ):
        self.part_code = part_code
        self.name = name
        self.dependencies = dependencies



"""Tests for automatic package discovery"""

import os
import shutil
import tempfile

import pytest

from setuptools import find_namespace_packages, find_packages
from setuptools.discovery import FlatLayoutPackageFinder

from .compat.py39 import os_helper


class TestFindPackages:
    def setup_method(self, method):
        self.dist_dir = tempfile.mkdtemp()
        self._make_pkg_structure()

    def teardown_method(self, method):
        shutil.rmtree(self.dist_dir)

    def _make_pkg_structure(self):
        """Make basic package structure.

        dist/
            docs/
                conf.py
            pkg/
                __pycache__/
                nspkg/
                    mod.py
                subpkg/
                    assets/
                        asset
                    __init__.py
            setup.py

        """
        self.docs_dir = self._mkdir('docs', self.dist_dir)
        self._touch('conf.py', self.docs_dir)
        self.pkg_dir = self._mkdir('pkg', self.dist_dir)
        self._mkdir('__pycache__', self.pkg_dir)
        self.ns_pkg_dir = self._mkdir('nspkg', self.pkg_dir)
        self._touch('mod.py', self.ns_pkg_dir)
        self.sub_pkg_dir = self._mkdir('subpkg', self.pkg_dir)
        self.asset_dir = self._mkdir('assets', self.sub_pkg_dir)
        self._touch('asset', self.asset_dir)
        self._touch('__init__.py', self.sub_pkg_dir)
        self._touch('setup.py', self.dist_dir)

    def _mkdir(self, path, parent_dir=None):
        if parent_dir:
            path = os.path.join(parent_dir, path)
        os.mkdir(path)
        return path

    def _touch(self, path, dir_=None):
        if dir_:
            path = os.path.join(dir_, path)
        open(path, 'wb').close()
        return path

    def test_regular_package(self):
        self._touch('__init__.py', self.pkg_dir)
        packages = find_packages(self.dist_dir)
        assert packages == ['pkg', 'pkg.subpkg']

    def test_exclude(self):
        self._touch('__init__.py', self.pkg_dir)
        packages = find_packages(self.dist_dir, exclude=('pkg.*',))
        assert packages == ['pkg']

    def test_exclude_recursive(self):
        """
        Excluding a parent package should not exclude child packages as well.
        """
        self._touch('__init__.py', self.pkg_dir)
        self._touch('__init__.py', self.sub_pkg_dir)
        packages = find_packages(self.dist_dir, exclude=('pkg',))
        assert packages == ['pkg.subpkg']

    def test_include_excludes_other(self):
        """
        If include is specified, other packages should be excluded.
        """
        self._touch('__init__.py', self.pkg_dir)
        alt_dir = self._mkdir('other_pkg', self.dist_dir)
        self._touch('__init__.py', alt_dir)
        packages = find_packages(self.dist_dir, include=['other_pkg'])
        assert packages == ['other_pkg']

    def test_dir_with_dot_is_skipped(self):
        shutil.rmtree(os.path.join(self.dist_dir, 'pkg/subpkg/assets'))
        data_dir = self._mkdir('some.data', self.pkg_dir)
        self._touch('__init__.py', data_dir)
        self._touch('file.dat', data_dir)
        packages = find_packages(self.dist_dir)
        assert 'pkg.some.data' not in packages

    def test_dir_with_packages_in_subdir_is_excluded(self):
        """
        Ensure that a package in a non-package such as build/pkg/__init__.py
        is excluded.
        """
        build_dir = self._mkdir('build', self.dist_dir)
        build_pkg_dir = self._mkdir('pkg', build_dir)
        self._touch('__init__.py', build_pkg_dir)
        packages = find_packages(self.dist_dir)
        assert 'build.pkg' not in packages

    @pytest.mark.skipif(not os_helper.can_symlink(), reason='Symlink support required')
    def test_symlinked_packages_are_included(self):
        """
        A symbolically-linked directory should be treated like any other
        directory when matched as a package.

        Create a link from lpkg -> pkg.
        """
        self._touch('__init__.py', self.pkg_dir)
        linked_pkg = os.path.join(self.dist_dir, 'lpkg')
        os.symlink('pkg', linked_pkg)
        assert os.path.isdir(linked_pkg)
        packages = find_packages(self.dist_dir)
        assert 'lpkg' in packages

    def _assert_packages(self, actual, expected):
        assert set(actual) == set(expected)

    def test_pep420_ns_package(self):
        packages = find_namespace_packages(
            self.dist_dir, include=['pkg*'], exclude=['pkg.subpkg.assets']
        )
        self._assert_packages(packages, ['pkg', 'pkg.nspkg', 'pkg.subpkg'])

    def test_pep420_ns_package_no_includes(self):
        packages = find_namespace_packages(self.dist_dir, exclude=['pkg.subpkg.assets'])
        self._assert_packages(packages, ['docs', 'pkg', 'pkg.nspkg', 'pkg.subpkg'])

    def test_pep420_ns_package_no_includes_or_excludes(self):
        packages = find_namespace_packages(self.dist_dir)
        expected = ['docs', 'pkg', 'pkg.nspkg', 'pkg.subpkg', 'pkg.subpkg.assets']
        self._assert_packages(packages, expected)

    def test_regular_package_with_nested_pep420_ns_packages(self):
        self._touch('__init__.py', self.pkg_dir)
        packages = find_namespace_packages(
            self.dist_dir, exclude=['docs', 'pkg.subpkg.assets']
        )
        self._assert_packages(packages, ['pkg', 'pkg.nspkg', 'pkg.subpkg'])

    def test_pep420_ns_package_no_non_package_dirs(self):
        shutil.rmtree(self.docs_dir)
        shutil.rmtree(os.path.join(self.dist_dir, 'pkg/subpkg/assets'))
        packages = find_namespace_packages(self.dist_dir)
        self._assert_packages(packages, ['pkg', 'pkg.nspkg', 'pkg.subpkg'])


class TestFlatLayoutPackageFinder:
    EXAMPLES = {
        "hidden-folders": (
            [".pkg/__init__.py", "pkg/__init__.py", "pkg/nested/file.txt"],
            ["pkg", "pkg.nested"],
        ),
        "private-packages": (
            ["_pkg/__init__.py", "pkg/_private/__init__.py"],
            ["pkg", "pkg._private"],
        ),
        "invalid-name": (
            ["invalid-pkg/__init__.py", "other.pkg/__init__.py", "yet,another/file.py"],
            [],
        ),
        "docs": (["pkg/__init__.py", "docs/conf.py", "docs/readme.rst"], ["pkg"]),
        "tests": (
            ["pkg/__init__.py", "tests/test_pkg.py", "tests/__init__.py"],
            ["pkg"],
        ),
        "examples": (
            [
                "pkg/__init__.py",
                "examples/__init__.py",
                "examples/file.py",
                "example/other_file.py",
                # Sub-packages should always be fine
                "pkg/example/__init__.py",
                "pkg/examples/__init__.py",
            ],
            ["pkg", "pkg.examples", "pkg.example"],
        ),
        "tool-specific": (
            [
                "htmlcov/index.html",
                "pkg/__init__.py",
                "tasks/__init__.py",
                "tasks/subpackage/__init__.py",
                "fabfile/__init__.py",
                "fabfile/subpackage/__init__.py",
                # Sub-packages should always be fine
                "pkg/tasks/__init__.py",
                "pkg/fabfile/__init__.py",
            ],
            ["pkg", "pkg.tasks", "pkg.fabfile"],
        ),
    }

    @pytest.mark.parametrize("example", EXAMPLES.keys())
    def test_unwanted_directories_not_included(self, tmp_path, example):
        files, expected_packages = self.EXAMPLES[example]
        ensure_files(tmp_path, files)
        found_packages = FlatLayoutPackageFinder.find(str(tmp_path))
        assert set(found_packages) == set(expected_packages)


def ensure_files(root_path, files):
    for file in files:
        path = root_path / file
        path.parent.mkdir(parents=True, exist_ok=True)
        path.touch()



import json
from abc import (
    ABCMeta,
    abstractmethod,
)
from typing import (
    Any,
    Dict,
)


class TableError(Exception):
    __metaclass__ = ABCMeta

    error_code = -1

    @property
    def error_message(self):
        # type: (...) -> str

        ex_dict = {
            "ex": self.__class__.__name__,
        }
        extra_data = self.get_extra_data()
        if extra_data is not None:
            ex_dict.update(extra_data)

        return json.dumps(ex_dict)

    @abstractmethod
    def get_extra_data(self):
        # type: (...) -> Dict[str, Any]

        raise NotImplementedError()



class ErrorFormattingReport(Exception):

    def __init__(
        self,
        report_name: str,
        error_message: str,
    ) -> None:
        super(ErrorFormattingReport, self).__init__(f"Error formatting report {report_name}: {error_message}")



from __future__ import annotations

import typing
from typing import Any
from typing import Mapping
from typing import Optional
from typing import Tuple

from .. import exc

if typing.TYPE_CHECKING:
    from .interfaces import _CoreAnyExecuteParams
    from .interfaces import _CoreMultiExecuteParams
    from .interfaces import _DBAPIAnyExecuteParams
    from .interfaces import _DBAPIMultiExecuteParams


_no_tuple: Tuple[Any, ...] = ()


def _distill_params_20(
    params: Optional[_CoreAnyExecuteParams],
) -> _CoreMultiExecuteParams:
    if params is None:
        return _no_tuple
    # Assume list is more likely than tuple
    elif isinstance(params, list) or isinstance(params, tuple):
        # collections_abc.MutableSequence): # avoid abc.__instancecheck__
        if params and not isinstance(params[0], (tuple, Mapping)):
            raise exc.ArgumentError(
                "List argument must consist only of tuples or dictionaries"
            )

        return params
    elif isinstance(params, dict) or isinstance(
        # only do immutabledict or abc.__instancecheck__ for Mapping after
        # we've checked for plain dictionaries and would otherwise raise
        params,
        Mapping,
    ):
        return [params]
    else:
        raise exc.ArgumentError("mapping or list expected for parameters")


def _distill_raw_params(
    params: Optional[_DBAPIAnyExecuteParams],
) -> _DBAPIMultiExecuteParams:
    if params is None:
        return _no_tuple
    elif isinstance(params, list):
        # collections_abc.MutableSequence): # avoid abc.__instancecheck__
        if params and not isinstance(params[0], (tuple, Mapping)):
            raise exc.ArgumentError(
                "List argument must consist only of tuples or dictionaries"
            )

        return params
    elif isinstance(params, (tuple, dict)) or isinstance(
        # only do abc.__instancecheck__ for Mapping after we've checked
        # for plain dictionaries and would otherwise raise
        params,
        Mapping,
    ):
        # cast("Union[List[Mapping[str, Any]], Tuple[Any, ...]]", [params])
        return [params]  # type: ignore
    else:
        raise exc.ArgumentError("mapping or sequence expected for parameters")



# -*- coding: utf-8 -*-

from typing import (
    Tuple,
    List,
    Dict,
)

from xml.etree import cElementTree as eTree
from xml.etree.ElementTree import Element

from msgbus import (
    TK_POS_LISTUSERS,
    FM_PARAM,
    TK_SYS_NAK,
)
from mw_helper import show_message_dialog
from old_helper import BaseRepository
from persistence import Connection
from sysactions import (
    UiOutOfSync,
    send_message,
    get_model,
    get_podtype,
    get_user_information,
    show_messagebox,
    StopAction,
    get_poslist,
)

from userhelper import UserHelper
from .delivery_functions import get_pos_id_of_remote_order
from utilfunctions import get_pos_id_from_session

from .. import (
    USECS_PER_SEC,
    logger,
)

mb_context = None


def list_opened_users(
    pos_id     # type: str
):
    # type: (...) -> List[Element]

    pos_id_format = _format_pos_id(pos_id=pos_id)
    msg = send_message(
        dest_name=pos_id_format,
        token=TK_POS_LISTUSERS,
        format=FM_PARAM,
        data="%s" % pos_id,
        timeout=600 * USECS_PER_SEC,
    )
    if msg.token == TK_SYS_NAK:
        logger.error("Logoff POS %d - Error listing logged operators" % int(pos_id))
        raise Exception("Error listing logged operators")

    model = get_model(pos_id)
    working_mode = model.find("WorkingMode").get("usrCtrlType")

    xml = eTree.XML(msg.data)
    opened_users = [tag for tag in xml.findall("User") if not tag.get("closeTime")]
    working_users = [tag for tag in opened_users if tag.get("userControlType") == working_mode]

    if working_mode == "QSR":
        working_users = [tag for tag in working_users if tag.get("posId") == pos_id]

    return working_users


def list_all_opened_users(
    pos_id,     # type: str
):
    # type: (...) -> List[Element]

    pos_id_format = _format_pos_id(pos_id=pos_id)
    msg = send_message(
        dest_name=pos_id_format,
        token=TK_POS_LISTUSERS,
        format=FM_PARAM,
        data="%s" % pos_id,
        timeout=600 * USECS_PER_SEC,
    )

    if msg.token == TK_SYS_NAK:
        logger.error("Logoff POS %d - Error listing logged operators" % int(pos_id))
        raise Exception("Error listing logged operators")

    xml = eTree.XML(msg.data)
    opened_users = [tag for tag in xml.findall("User") if not tag.get("closeTime")]
    working_users = [tag for tag in opened_users]

    return working_users


def list_ids_opened_users():
    with UserHelper(mb_context) as user_helper:
        users_id = []
        users = user_helper.get_all_logged_users()
        for x in users:
            users_id.append(int(x))

        return users_id


def check_valid_user(pos_id, user_id):
    user_xml = get_user_information(user_id)
    if user_xml is None:
        show_message_dialog(pos_id, message="$INVALID_USER_OR_PASSWORD")
        return False

    user_list = eTree.XML(user_xml).findall("user")
    user_filter = filter(lambda x: x.get("UserId") == str(user_id), user_list)
    if user_filter in [None, [], ""]:
        show_message_dialog(pos_id, message="$INVALID_USER_OR_PASSWORD")
        return False

    return True


def get_opened_users():
    # type: (...) -> List[Tuple[int, Element]]

    is_ts_listed = False
    pos_ids_to_ignore = [get_pos_id_of_remote_order()]
    pos_user_list = []
    for pos_id_to_list in sorted(get_poslist()):
        if pos_id_to_list in pos_ids_to_ignore:
            continue

        opened_users = list_opened_users(pos_id=str(pos_id_to_list))
        if len(opened_users) == 0:
            continue

        opened_users.sort(key=lambda x: x.get("id"))
        model = get_model(posid=pos_id_to_list)
        pod_type = get_podtype(model=model)
        if pod_type == "TT":
            pos_ids_to_ignore.append(pos_id_to_list)
            continue

        working_mode = model.find("WorkingMode").get("usrCtrlType")
        if working_mode == "TS":
            if is_ts_listed:
                continue

            is_ts_listed = True

        for user in opened_users:
            user_pos = int(user.get("posId", pos_id_to_list))
            pos_user_list.append((user_pos, user))

    return pos_user_list


def is_user_already_logged(
    pos_id,           # type: str
    user_id,          # type: str
    try_fix=True,     # type: bool
):
    # type: (...) -> bool

    params = "\0".join(["%s" % pos_id, "true"])

    pos_id_format = _format_pos_id(pos_id=pos_id)
    msg = send_message(
        dest_name=pos_id_format,
        token=TK_POS_LISTUSERS,
        format=FM_PARAM,
        data=params,
        timeout=600 * USECS_PER_SEC,
    )
    if msg.token == TK_SYS_NAK:
        logger.error("Logoff POS %d - Error getting logged users" % int(pos_id))
        return True

    xml = eTree.XML(msg.data)
    opened_users = [tag for tag in xml.findall("User") if not tag.get("closeTime")]
    working_users = [tag for tag in opened_users if tag.get("id") == user_id]

    if len(working_users) > 0:
        model = get_model(pos_id)
        _user_has_same_working_mode(model, opened_users, pos_id, user_id)

        user_info = [x for x in opened_users if x.get("id") == user_id][0]
        operator = [op for op in model.findall("Operators/Operator") if op.get("id") == user_id]

        if _has_opened_user(
            operator=operator,
            user_info=user_info,
            pos_id=pos_id,
        ):
            if operator[0].get("state") != "LOGGEDIN":
                return False

            user_logged_in_current_pos = user_info.get("posId") == str(pos_id)
            if user_logged_in_current_pos:
                logger.info("Send event update model ui. POS {} - USER {}".format(pos_id, user_id))
                raise UiOutOfSync(
                    pos_id=pos_id,
                    event_data=user_id,
                )

        old_login = model.find("PosState").get("period") != user_info.get("sessionId").split(",period=")[-1]
        if try_fix and old_login:
            _fix_opened_users(pos_id, user_id)
            return is_user_already_logged(pos_id, user_id, False)

        _show_already_logged_message(
            pos_id=pos_id,
            user_info=user_info,
        )
        return True

    return False


def _has_opened_user(
    operator,    # type: List[Dict[str]]
    user_info,   # type: Dict[str]
    pos_id,      # type: str
):
    # type: (...) -> bool

    user_is_logged_in_other_pos = _is_logged_in_other_qsr_pos(
        user_info=user_info,
        pos_id=pos_id,
    )
    return len(operator) > 0 and not user_is_logged_in_other_pos


def _format_pos_id(
    pos_id,    # type: str
):
    # type: (...) -> str

    return "POS%d" % int(pos_id)


def _show_already_logged_message(
    pos_id,        # type: str
    user_info,     # type: Dict[str]
):
    # type: (...) -> None

    session_id = user_info.get("sessionId")
    logged_pos_id = get_pos_id_from_session(session_id=session_id) or user_info.get("posId")
    show_message_dialog(
        pos_id=pos_id,
        message="$OPERATOR_ALREADY_LOGGED|{}".format(logged_pos_id),
    )


def _user_has_same_working_mode(model, opened_users, pos_id, user_id):
    user_info = [x for x in opened_users if x.get("id") == user_id]
    if user_info:
        pos_working_mode = model.find("WorkingMode").get("usrCtrlType")
        operator_working_mode = "QSR" if "pos" in user_info[0].get("sessionId") else "TS"
        if pos_working_mode != operator_working_mode:
            show_messagebox(pos_id, "$USER_LOGGED_IN_ANOTHER_POS_TYPE|{}".format(user_info[0].get("posId")))
            raise StopAction()


def _fix_opened_users(pos_id, user_id):
    # type: (str, str) -> None
    def inner_func(conn):
        # type: (Connection) -> None
        query = _fix_opened_users_query.format(user_id)
        conn.query(query)

    repository = BaseRepository(mb_context)
    repository.execute_with_connection(inner_func, pos_id)


def _is_logged_in_other_qsr_pos(
    user_info,    # type: str
    pos_id,       # type: str
):
    # type: (...) -> bool

    user_control_type = user_info.get("userControlType")
    if user_control_type != "QSR":
        return False

    session_id = user_info.get("sessionId")
    logged_pos_id = get_pos_id_from_session(session_id=session_id)
    return "closeTime" not in user_info and pos_id != logged_pos_id


def set_context(context):
    global mb_context
    mb_context = context


_fix_opened_users_query = """
    UPDATE UserCount SET CloseTime = (
        SELECT MAX(OpenTime) FROM UserCount WHERE OperatorId = '{0}'
    ) WHERE OperatorId = '{0}' AND CloseTime is NULL AND OpenTime < (
        SELECT MAX(CloseTime) FROM UserCount WHERE OperatorId = '{0}'
    )
"""



from enum import IntEnum

from messagebus import (
    TokenCreator,
    TokenPriority,
)

MSGGRP_TK_POS_UPDATER = "95"


def _create_token(
    number,     # type: str
):
    # type: (...) -> int

    return TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group=MSGGRP_TK_POS_UPDATER,
        number=number,
    )


# noinspection SpellCheckingInspection

class PosUpdaterApiTokens(IntEnum):

    TK_POS_UPDATER_PERFORM_CATALOG_UPDATE = _create_token("1")
    TK_POS_UPDATER_PERFORM_LOADER_UPDATE = _create_token("5")
    TK_POS_UPDATER_PERFORM_BINARIES_UPDATE = _create_token("7")
    TK_POS_UPDATER_PERFORM_CHROME_UPDATE = _create_token("9")



import logging

from sysactions import (
    action,
    close_asynch_dialog,
    get_model,
    get_posfunction,
    show_confirmation,
    show_messagebox,
)
from xml.etree import cElementTree as eTree

from actions.login.unpause_user import unpause_user
from actions.table.create_table_order import create_table_order
from api.table import get_table_service
from order_api.dto import OrderState
from table_manager_api import TableService
from table_manager_api.dto import (
    TableServiceDto,
    TableStatus,
)
from table_manager_api.dto.exception import (
    InvalidTabMask,
    InvalidTabName,
    TabNotFound,
    TabNotInValidTabNames,
)
from thread_local import get_current_model
from typing import (
    Optional,
    Tuple,
)

logger = logging.getLogger("TableActions")


@action
def get_tab_info(
    pos_id,     # type: str
    tab_name,   # type: str
):
    # type: (...) -> Optional[str]

    pos_id = int(pos_id)
    table_service = get_table_service(pos_id=pos_id)
    try:
        tab_service_dto = table_service.get_service_tab(tab_name=tab_name)
        return eTree.tostring(tab_service_dto.to_xml())
    except (Exception,):
        logger.info("Cannot obtain existing service for tab '{}'".format(tab_name))
        show_messagebox(
            posid=pos_id,
            message="$TAB_DOES_NOT_HAVE_STARTED_SERVICE",
        )
        return None


@action
def read_tab_name(
    pos_id,         # type: str
    name,           # type: str
    user_id=None,   # type: Optional[str]
):
    # type: (...) -> bool

    if user_id:
        logged_user = unpause_user(
            pos_id=pos_id,
            user_id=user_id,
        )
        if not logged_user:
            return False

    pos_id = int(pos_id)
    pos_function = get_pos_function()
    table_service = get_table_service(pos_id=pos_id)

    tab_service_dto, stop_process = get_tab_by_name(
        pos_id=pos_id,
        name=name,
        pos_function=pos_function,
        table_service=table_service
    )
    if stop_process:
        return True

    if tab_service_dto is None:
        tab_service_dto = table_service.start_tab(tab_name=name)
        return _execute_action_for_new_tab(
            pos_id=pos_id,
            pos_function=pos_function,
            tab_service_dto=tab_service_dto,
        )

    opened_in_other_pos, tab_service_dto = _tab_is_opened_in_other_pos(
        pos_id=pos_id,
        tab_service_dto=tab_service_dto,
        table_service=table_service,
    )
    if opened_in_other_pos:
        return False

    return _execute_action_for_existing_tab(
        pos_id=pos_id,
        table_service=table_service,
        pos_function=pos_function,
        tab_service_dto=tab_service_dto,
    )


def get_pos_function():
    # type: (...) -> str

    model = get_current_model()
    return get_posfunction(model=model)


def get_tab_by_name(
    pos_id,         # type: int
    name,           # type: str
    pos_function,   # type: str
    table_service,  # type: TableService
):
    # type: (...) -> Tuple[Optional[TableServiceDto], bool]

    tab_service_dto = None
    stop_process = False

    try:
        tab_service_dto = table_service.get_service_tab(tab_name=name)
    except (InvalidTabMask, TabNotInValidTabNames) as ex:
        show_messagebox(
            posid=pos_id,
            message=ex.message.split("|")[0],
        )
        stop_process = True

    except TabNotFound:
        if pos_function == "CS":
            tab_service_dto = _create_order_in_pos_cs(
                pos_id=pos_id,
                table_service=table_service,
                tab_name=name,
            )
            stop_process = True

    except InvalidTabName:
        show_messagebox(
            posid=pos_id,
            message="$EXISTS_AS_A_TABLE",
        )
        stop_process = True

    return tab_service_dto, stop_process


def _create_order_in_pos_cs(
    pos_id,                 # type: int
    table_service,          # type: TableService
    tab_name=None,          # type: Optional[str]
    tab_service_dto=None,   # type: Optional[TableServiceDto]
):
    # type: (...) -> Optional[TableServiceDto]

    from table_actions import (
        abandon_table,
        create_table_modified_event,
    )

    proceed = show_confirmation(
        posid=pos_id,
        message="$EMPTY_TAB_WANT_TO_ADD_ITEMS",
    )
    if not proceed:
        if tab_service_dto:
            abandon_table(
                pos_id=pos_id,
                table_id=tab_service_dto.table.table_id,
            )
        return None

    if not tab_service_dto:
        tab_service_dto = table_service.start_tab(tab_name=tab_name)
        create_table_modified_event(
            pos_id=pos_id,
            table_picture=eTree.tostring(tab_service_dto.to_xml()),
        )

    create_table_order(
        pos_id=pos_id,
        table_id=tab_service_dto.table.table_id,
    )

    return tab_service_dto


def _execute_action_for_new_tab(
    pos_id,             # type: int
    pos_function,       # type: str
    tab_service_dto,    # type: TableServiceDto
):
    # type: (...) -> bool

    if pos_function == "OT":
        create_table_order(
            pos_id=pos_id,
            table_id=tab_service_dto.table.table_id,
        )

    from table_actions import create_table_modified_event

    create_table_modified_event(
        pos_id=pos_id,
        table_picture=eTree.tostring(tab_service_dto.to_xml()),
    )

    return True


def _execute_action_for_existing_tab(
    pos_id,             # type: int
    table_service,      # type: TableService
    pos_function,       # type: str
    tab_service_dto,    # type: TableServiceDto
):
    # type: (...) -> bool

    if pos_function == "OT":
        return _open_tab_in_order_taker_mode(
            pos_id=pos_id,
            tab_service_dto=tab_service_dto,
        )

    if pos_function == "CS":
        return _open_tab_in_cashier_mode(
            pos_id=pos_id,
            table_service=table_service,
            tab_service_dto=tab_service_dto,
        )
    if pos_function == "FL":
        _open_tab_in_full_mode(
            pos_id=pos_id,
            tab_service_dto=tab_service_dto,
        )

    return True


def _tab_is_opened_in_other_pos(
    pos_id,             # type: int
    tab_service_dto,    # type: TableServiceDto
    table_service,      # type: TableService
):
    # type: (...) -> Tuple[bool, Optional[TableServiceDto]]

    updated_tab_service_dto = None
    tab_in_another_pos = True
    if tab_service_dto.parent_table_id:
        show_messagebox(
            posid=pos_id,
            message="$COULD_NOT_SELECT_TAB_BECAUSE_IS_LINKED_TO_ANOTHER_TAB|{}".format(tab_service_dto.table.name),
        )
        return tab_in_another_pos, updated_tab_service_dto

    tab_pos_id = _get_table_pos_id(table_service_dto=tab_service_dto)
    if tab_pos_id and tab_pos_id != pos_id:
        success = _close_tab_in_other_pos(
            pos_id=pos_id,
            tab_pos_id=tab_pos_id,
        )
        if not success:
            return tab_in_another_pos, updated_tab_service_dto

        updated_tab_service_dto = table_service.get_service(table_id=tab_service_dto.table.table_id)

    tab_in_another_pos = False
    updated_tab_service_dto = updated_tab_service_dto or tab_service_dto
    return tab_in_another_pos, updated_tab_service_dto


def _close_tab_in_other_pos(
    pos_id,         # type: int
    tab_pos_id,     # type: int
):
    # type: (...) -> bool

    proceed = show_confirmation(
        posid=pos_id,
        message="$TAB_OPENED_IN_ANOTHER_POS_DO_YOU_WANT_TO_RECOVER_HERE|{}".format(tab_pos_id),
    )
    if not proceed:
        return False

    wait_dialog_id = show_messagebox(
        posid=pos_id,
        message="$WAITING_FOR_OPERATOR_CONFIRMATION",
        buttons="",
        asynch=True,
    )

    proceed = show_confirmation(
        posid=tab_pos_id,
        message="$ALLOWS_RECOVER_TAB_NON_KITCHEN_ORDERS_WILL_BE_DELETED",
        icon="warning",
        timeout=10000,
    )

    close_asynch_dialog(
        posid=pos_id,
        dlgid=wait_dialog_id,
    )
    if proceed is False:
        show_messagebox(
            posid=pos_id,
            message="$OPERATION_CANCELLED_BY_OPERATOR",
            timeout=5000,
        )
        return False

    from actions.login.utils import close_current_order

    closed_order = close_current_order(
        model=get_model(posid=tab_pos_id),
        current_pos_id=str(pos_id),
        pos_to_close_order=str(tab_pos_id),
    )
    if not closed_order:
        return False

    return True


def _get_table_pos_id(
    table_service_dto,  # type: TableServiceDto
):
    # type: (...) -> Optional[int]

    if table_service_dto.table.pos_id is not None:
        return table_service_dto.table.pos_id

    for order in table_service_dto.service_orders:
        if order.state == OrderState.IN_PROGRESS:
            return order.order_xml.get("posId")

    return None


def _open_tab_in_full_mode(
    pos_id,             # type: int
    tab_service_dto,    # type: TableServiceDto
):
    # type: (...) -> None

    from actions.table import reopen_table
    from table_actions import select_table

    if tab_service_dto.status == TableStatus.Totaled:
        reopen_table(
            pos_id=pos_id,
            table_id=tab_service_dto.table.table_id,
        )

        return

    select_table(
        pos_id=pos_id,
        table_id=tab_service_dto.table.table_id,
    )


def _open_tab_in_order_taker_mode(
    pos_id,             # type: int
    tab_service_dto,    # type: TableServiceDto
):
    # type: (...) -> bool

    _open_tab_in_full_mode(
        pos_id=pos_id,
        tab_service_dto=tab_service_dto,
    )

    create_table_order(
        pos_id=pos_id,
        table_id=tab_service_dto.table.table_id,
    )
    return True


def _open_tab_in_cashier_mode(
    pos_id,             # type: int
    table_service,      # type: TableService
    tab_service_dto,    # type: TableServiceDto
):
    # type: (...) -> bool

    from table_actions import (
        abandon_table,
        do_total_table,
        select_table,
    )
    from actions.table import deselect_table

    success = select_table(
        pos_id=pos_id,
        table_id=tab_service_dto.table.table_id,
    )
    if not success:
        return False

    if tab_service_dto.status == TableStatus.Totaled:
        return True

    if not _has_sent_kitchen_orders(tab_service_dto=tab_service_dto):
        _create_order_in_pos_cs(
            pos_id=pos_id,
            table_service=table_service,
            tab_service_dto=tab_service_dto,
        )
        return True

    try:
        do_total_table(
            pos_id=pos_id,
            table_id=tab_service_dto.table.table_id,
            display_error=False,
        )
        return True

    except Exception as ex:
        if "$COULD_NOT_TOTALIZE_EMPTY_TABLE" in str(ex) or "$INVALID_TABLE_STATUS" in str(ex):
            logger.info("Could not totalize empty tab {}".format(tab_service_dto.table.table_id))
            proceed = show_confirmation(
                posid=pos_id,
                message="$WANT_TO_CLOSE_TAB_BECAUSE_IS_OPENED_BUT_HAS_NOT_ORDERS",
            )
            if proceed:
                return abandon_table(
                    pos_id=pos_id,
                    table_id=tab_service_dto.table.table_id,
                )

        deselect_table(pos_id=pos_id)

    return False


def _has_sent_kitchen_orders(
    tab_service_dto,    # type: TableServiceDto
):
    # type: (...) -> bool

    for order in tab_service_dto.service_orders:
        if order.state == OrderState.STORED:
            state_history = [state.state for state in order.state_history]
            if OrderState.TOTALED not in state_history:
                return True

    return False



import json
import logging

from helper import remove_accents
from mw_helper import (
    show_keyboard,
    show_message_dialog,
    show_tab_customer_dialog,
)
from sysactions import (
    action,
    show_messagebox,
)
from api.table import get_table_service
from application.interactor import (
    StartServiceInteractor,
    TabsInteractor,
)
from application.interactor.service.impl import TableReleaserImpl
from custom_sysactions import (
    block_action_if_pos_is_blocked,
    user_must_be_logged,
    user_authorized_to_execute_action,
)
from msgbusboundary import MsgBusTableService as tableService
from msgbusboundary.ui import MsgBusScreenChanger
from table_manager_api.dto.exception import InvalidTabMask
from typing import (
    Dict,
    Optional,
    Union,
)
from utilfunctions import is_valid_string

logger = logging.getLogger("TableActions")


@action
def show_open_tabs(
    pos_id,     # type: int
):
    # type: (...) -> bool

    tabs_interactor = TabsInteractor(int(pos_id))
    opened_tabs = tabs_interactor.get_opened_tabs()
    if not opened_tabs:
        show_messagebox(pos_id, "$EMPTY_OPENED_TABS")
        return False

    tabs_interactor.show_opened_tabs(opened_tabs)

    return True


@action
@block_action_if_pos_is_blocked
@user_must_be_logged
@user_authorized_to_execute_action
def create_tab(
    pos_id,         # type: Union[str, int]
    tab_info=None,  # type: Optional[Dict[str, Union[str, int]]]
):

    return open_tab(
        pos_id=pos_id,
        tab_info=tab_info,
    )


@action
def open_tab(
    pos_id,                 # type: Union[str, int]
    parent_table_id=None,   # type: Optional[str]
    tab_info=None,          # type: Optional[Dict[str, Union[str, int]]]
):
    # type: (...) -> Union[str, bool]

    from table_actions import (
        allowed_to_create_table,
        create_table_modified_event,
    )
    from select_tab_by_barcode import (
        get_tab_by_name,
        get_pos_function,
    )

    if parent_table_id is None and not allowed_to_create_table(
        pos_id=pos_id,
        table_id="",
        is_tab=True,
    ):
        show_message_dialog(
            pos_id=pos_id,
            message="$OPERATOR_DOES_NOT_BELONG_TO_TABLE_SECTOR",
        )
        return False

    if not tab_info:
        tab_info = get_tab_info(pos_id=pos_id)
    if not tab_info:
        return False

    tab_name = tab_info.pop("TAB_NAME", None)
    if not tab_name:
        return False

    tab_service_dto, stop_process = get_tab_by_name(
        pos_id=pos_id,
        name=tab_name,
        pos_function=get_pos_function(),
        table_service=get_table_service(pos_id=pos_id),
    )
    if stop_process:
        return False

    if tab_service_dto:
        show_messagebox(
            posid=pos_id,
            message="$TAB_ALREADY_EXISTS",
        )
        logger.warning("Tab {} already exists".format(tab_name))
        return False

    table_releaser = TableReleaserImpl(table_service=tableService())
    screen_changer = MsgBusScreenChanger()

    interactor = StartServiceInteractor(
        screen_changer=screen_changer,
        table_service=tableService(),
        table_releaser=table_releaser,
    )
    parent_table_id = parent_table_id or ""
    tab_picture = None
    try:
        tab_picture = interactor.execute(
            pos_id=pos_id,
            name=tab_name,
            source_table_id=parent_table_id,
            custom_properties=tab_info,
        )
    except InvalidTabMask:
        show_messagebox(
            posid=pos_id,
            message="$TAB_NOT_FOUND",
        )
    except (Exception, ):
        logger.exception("Unmapped error")
        raise

    create_table_modified_event(
        pos_id=pos_id,
        table_picture=tab_picture,
    )

    return tab_picture


def get_tab_info(
    pos_id,     # type: str
):
    # type: (...) -> Optional[Dict[str, str]]

    message = None

    while True:
        if message:
            show_message_dialog(
                pos_id=pos_id,
                message=message,
            )

        customer_data = _get_tab_input(pos_id=pos_id)
        if not customer_data:
            return None

        tab_name = customer_data["TAB_NAME"]
        if tab_name is None:
            return None

        message = _get_invalid_data_message(
            customer_data=customer_data,
            tab_name=tab_name,
        )
        if message:
            continue

        break
    return customer_data


def _get_tab_input(
    pos_id,         # type: str
):
    # type: (...) -> Optional[Dict[str, str]]

    from table_actions import customer_info_tab

    if customer_info_tab:
        customer = show_tab_customer_dialog(pos_id=pos_id)
        if customer is None:
            return None

        customer_data = _format_customer_data(customer=customer)
    else:
        customer_data = {"TAB_NAME": None}
        tab_name = show_keyboard(
            posid=pos_id,
            message="$TYPE_TAB_NAME",
            defvalue="",
            title="",
            numpad=False,
            buttons="$OK|$PREVIOUS",
        )
        if tab_name is not None:
            customer_data["TAB_NAME"] = remove_accents(text=tab_name)

    return customer_data


def _are_all_values_filled(
    obj_dict,   # type: Dict[str, Optional[str]]
):
    # type: (...) -> bool

    return all([False for value in obj_dict.values() if len(value) == 0])


def _get_invalid_data_message(
    customer_data,  # type: Dict[str, str]
    tab_name,       # type: str
):
    # type: (...) -> str

    from table_actions import (
        customer_info_tab,
        max_number_of_tab_digits,
    )

    has_all_values = _are_all_values_filled(obj_dict=customer_data)
    if not has_all_values:
        return "$INVALID_COSTUMER_INFO"

    if not is_valid_string(string=tab_name) or len(tab_name) > max_number_of_tab_digits:
        return "$INVALID_TAB_NAME|{}".format(max_number_of_tab_digits)

    if "customer_phone" in customer_info_tab and len(customer_data["CUSTOMER_PHONE"]) < 10:
        return "$INVALID_CUSTOMER_PHONE"

    if "customer_name" in customer_info_tab and not is_valid_string(string=customer_data["CUSTOMER_NAME"]):
        return "$INVALID_CUSTOMER_NAME"

    return ""


def _format_customer_data(
    customer,   # type: str
):
    # type: (...) -> Dict[str, str]

    from table_actions import customer_info_tab

    customer_data = json.loads(customer)
    if "customer_name" in customer_info_tab:
        customer_data["CUSTOMER_NAME"] = remove_accents(customer_data["CUSTOMER_NAME"])

    customer_data["TAB_NAME"] = remove_accents(customer_data["TABLE_NAME"])

    return customer_data



from datetime import datetime

from xml.etree import cElementTree as eTree

from model.log import LogEntry
from processor.event import EventProcessor


def _get_log_entry_from_pos_event(
    data,           # type: str
    event_type,     # type: str
):
    # type: (...) -> LogEntry

    event = eTree.fromstring(data)
    event_data = event[0] if "Event" in event.tag else event
    utc_now = datetime.utcnow()
    created_at = utc_now.strftime("%Y-%m-%d %H:%M:%S")

    return LogEntry(
        pos_id=event_data.get("posId", event_data.get("posid")) or -1,
        business_period=event_data.get("businessPeriod", utc_now.strftime("%Y%m%d")),
        event_type=event_type,
        created_at_gmt=created_at,
        session_id=event_data.get("sessionId"),
        event_data=data,
    )


class GenericEventProcessor(EventProcessor):

    def parse_data(
        self,           # type: GenericEventProcessor
        data,           # type: str
        event_type,     # type: str
    ):
        # type: (...) -> LogEntry

        return _get_log_entry_from_pos_event(
            data=data,
            event_type=event_type,
        )



import json

from update_api.dto import PosUpdaterDto


class BinariesUpdaterDto(PosUpdaterDto):

    def __init__(
        self,           # type: BinariesUpdaterDto
        pos_id,         # type: int
        user_id,        # type: str
        force_update,   # type: bool
    ):
        # type: (...) -> None

        super(BinariesUpdaterDto, self).__init__(
            pos_id=pos_id,
            user_id=user_id,
        )

        self.force_update = force_update

    def __str__(
        self,  # type: BinariesUpdaterDto
    ):
        # type: (...) -> str

        return json.dumps({
            "pos_id": self.pos_id,
            "user_id": self.user_id,
            "force_update": self.force_update,
        })



from ._FireHoldPriorityOrderSplitter import FireHoldPriorityOrderSplitter
from ._PrepTimeOrderSplitter import PrepTimeOrderSplitter
from ._SaleLinesOrderSplitter import SaleLinesOrderSplitter
from ._StoredOrderSplitter import StoredOrderSplitter



from messagebus import (
    TokenPriority,
    TokenCreator,
)

nreports_group = "6F"


TK_NREPORTS_BUILD_REPORT = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=nreports_group,
    number="1",
)  # 116391937 // 0x6F00001

TK_NREPORTS_REPORT_JSON_PARSER = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=nreports_group,
    number="2",
)  # 116391938 // 0x6F00002



from production.box import OrderStateTimerBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class OrderStateTimerBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(OrderStateTimerBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> OrderStateTimerBox
        return OrderStateTimerBox(config.name,
                                  config.sons,
                                  self.get_extra(config, "OrderState", []),
                                  self.get_logger(config))



import json
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    List,
    Optional,
)
from xml.etree import cElementTree as eTree

import iso8601
from nreports.model import ProductTypeName
from nreports.repository import ProductRepository


@dataclass
class Tender:
    value: str = ""
    tender_type: str = ""
    tender_type_description: str = ""
    tender_change_amount: str = ""


@dataclass
class Item:
    part_code: Optional[int] = None
    product_name: str = ""
    item_type: Optional[str] = None
    quantity: Optional[int] = None
    price: Optional[float] = None
    observation: Optional[str] = None
    is_cfh: bool = False


class DeliveryOrderDto:

    def __init__(
        self,
        order_picture_xml: eTree.Element,
        delivery_json: Dict[str, Any],
        product_repository: ProductRepository,
    ) -> None:

        self.product_repository = product_repository
        self.order_number = delivery_json.get("shortReference", "0")
        self.order_date = iso8601.parse_date(delivery_json.get("createAt", "2099-01-01T00:00:00.00000Z"))
        self.items = self._get_items(order_items=delivery_json.get("items", []))
        self.total_price = delivery_json.get("totalPrice", "0.00")
        self.sub_total_price = delivery_json.get("subTotal", "0.00")
        self.delivery_fee = float(delivery_json.get("deliveryFee", "0.00"))
        self.observation = self._get_custom_property(
            delivery_json=delivery_json,
            key="observation",
        )
        self.observations = delivery_json.get("observations", "")
        self.tenders = self._build_tenders(
            order_picture_xml=order_picture_xml,
            delivery_json=delivery_json,
        )
        self.customer_name = self._get_custom_property(
            delivery_json=delivery_json,
            key="customer_name",
        )
        self.customer_doc = self._get_custom_property(
            delivery_json=delivery_json,
            key="customer_doc",
        )
        self.customer_phone = self._get_custom_property(
            delivery_json=delivery_json,
            key="customer_phone",
        )
        phone_localizer = self._get_custom_property(
            delivery_json=delivery_json,
            key="customer_phone_localizer",
        )
        if phone_localizer:
            self.customer_phone += " ID: {}".format(phone_localizer)

        self.additional_fee = float(
            self._get_custom_property(
                delivery_json=delivery_json,
                key="additional_fee"
            ) or 0
        )
        self.pos_order_id = order_picture_xml.get("orderId")[-4:]
        self.partner = delivery_json["partner"]
        self.display_partner = self._get_custom_property(
            delivery_json=delivery_json,
            key="display_partner",
        )
        self.need_logistics = delivery_json.get("needLogistics", False)

        address = delivery_json.get("pickup", {}).get("address") or {}

        self.address = address.get("formattedAddress")
        self.address_neighborhood = address.get("neighborhood")
        self.city = address.get("city")
        self.address_complement = address.get("complement")
        self.postal_code = address.get("postalCode")
        self.reference = address.get("reference")
        self.take_out_order = delivery_json.get("pickup", {}).get("type", "").lower() == "take_out"

        self.vouchers = {}
        self._create_vouchers(delivery_json=delivery_json)

    def _get_items(
        self,
        order_items: List[Dict[str, Any]],
        father_qty: int = 1,
        level: int = 0,
    ) -> List[Item]:

        items = []
        for order_item in order_items:
            item = self._convert_order_item_into_item(order_item=order_item)

            if item.product_name is None:
                continue

            if item.quantity is not None:
                item.quantity *= father_qty

            if item.item_type != ProductTypeName.OPTION and not item.is_cfh:
                item.product_name = "  " * level + item.product_name
                items.append(item)

                if item.observation:
                    items.append(Item(product_name=item.observation))

            if "parts" in order_item:
                current_level = level if item.item_type == ProductTypeName.OPTION else level + 1
                items.extend(
                    self._get_items(
                        order_items=order_item.get("parts", []),
                        father_qty=item.quantity,
                        level=current_level,
                    )
                )

        return items

    def _convert_order_item_into_item(
        self,
        order_item: Dict[str, Any],
    ) -> Item:

        part_code = int(order_item.get("partCode"))
        observation = order_item.get("observation") or ""
        product_data = self.product_repository.get_product_data(product_id=part_code)
        return Item(
            part_code=part_code,
            product_name=product_data.product_name if product_data else None,
            item_type=order_item.get("itemType") or (product_data.product_type if product_data else ""),
            quantity=order_item.get("quantity"),
            price=order_item.get("price"),
            observation="   " + observation.replace("\n", " ") if observation else None,
            is_cfh=product_data.product_cfh if product_data else False,
        )

    def _create_vouchers(
        self,
        delivery_json: Dict[str, Any],
    ) -> None:

        self.vouchers["ORDER"] = 0
        self.vouchers["TOTAL_ITEMS"] = 0
        self.vouchers["DELIVERY_FEE"] = 0

        if "vouchers" in delivery_json and "store" in delivery_json["vouchers"]:
            self._get_order_vouchers(delivery_json=delivery_json)
            self._get_total_items_vouchers(delivery_json=delivery_json)
            self._get_delivery_fee_vouchers(delivery_json=delivery_json)

    def _get_order_vouchers(
        self,
        delivery_json: Dict[str, Any],
    ) -> None:

        if "order" in delivery_json["vouchers"]["store"]:
            self.vouchers["ORDER"] += delivery_json["vouchers"]["store"]["order"]

        if "order" in delivery_json["vouchers"]["partner"]:
            self.vouchers["ORDER"] += delivery_json["vouchers"]["partner"]["order"]

    def _get_total_items_vouchers(
        self,
        delivery_json: Dict[str, Any],
    ) -> None:

        if "total_items" in delivery_json["vouchers"]["store"]:
            self.vouchers["TOTAL_ITEMS"] += delivery_json["vouchers"]["store"]["total_items"]

        if "total_items" in delivery_json["vouchers"]["partner"]:
            self.vouchers["TOTAL_ITEMS"] += delivery_json["vouchers"]["partner"]["total_items"]

    def _get_delivery_fee_vouchers(
        self,
        delivery_json: Dict[str, Any],
    ) -> None:

        if "delivery_fee" in delivery_json["vouchers"]["store"]:
            self.vouchers["DELIVERY_FEE"] += delivery_json["vouchers"]["store"]["delivery_fee"]

        if "delivery_fee" in delivery_json["vouchers"]["partner"]:
            self.vouchers["DELIVERY_FEE"] += delivery_json["vouchers"]["partner"]["delivery_fee"]

    @staticmethod
    def _get_custom_property(
        delivery_json: Dict[str, Any],
        key: str,
    ) -> str:

        if "custom_properties" in delivery_json:
            value = [x for x in delivery_json["custom_properties"] if x["key"].lower() == key.lower()]
            if len(value) > 0:
                return value[0]["value"]

        return ""

    def _build_tenders(
        self,
        order_picture_xml: eTree.Element,
        delivery_json: Dict[str, Any],
    ) -> List[Tender]:

        delivery_partner = self._get_custom_property(
            delivery_json=delivery_json,
            key="partner",
        )
        if delivery_partner == "MANUAL":
            return self._get_tenders_manual_delivery_json(order_picture_xml=order_picture_xml)

        return self._get_tenders_from_delivery_json(delivery_json=delivery_json)

    @staticmethod
    def _get_tenders_manual_delivery_json(
        order_picture_xml: eTree.Element,
    ) -> List[Tender]:

        tender_list = []
        for tender_element in order_picture_xml.findall(".//TenderHistory/Tender"):
            if float(tender_element.get("tenderAmount")) <= 0:
                continue

            tender_detail = tender_element.get("tenderDetail")
            tender_detail_json = json.loads(tender_detail) if tender_detail else {}
            tender_data = Tender(
                value=tender_element.get("tenderAmount", ""),
                tender_type=tender_element.get("tenderDescr", ""),
                tender_type_description=tender_detail_json.get("tenderTypeDescription", ""),
                tender_change_amount=tender_detail_json.get("tenderChangeAmount", "")
            )
            tender_list.append(tender_data)

        return tender_list

    @staticmethod
    def _get_tenders_from_delivery_json(
        delivery_json: Dict[str, Any],
    ) -> List[Tender]:

        tenders = delivery_json.get("tenders", [])
        tender_list = []
        for tender in tenders:
            if float(tender.get("value", 0)) <= 0:
                continue

            tender_data = Tender(
                value=tender.get("value", ""),
                tender_type=tender.get("type", ""),
            )
            tender_list.append(tender_data)

        return tender_list



from logging import Logger

from production.model import Item
from production.model import ProductionOrder
from typing import Dict, List, Optional
from ._OrderChangerProductionBox import OrderChangerProductionBox
from ._ProductionBox import ProductionBox


class OrderCostAdderBox(OrderChangerProductionBox):
    def __init__(
        self,               # type: OrderCostAdderBox
        name,               # type: str
        sons,               # type: List[ProductionBox]
        per_item_cost,      # type: Dict[int, int]
        default_cost,       # type: int
        logger=None,        # type: Optional[Logger]
    ):
        super(OrderCostAdderBox, self).__init__(name, sons, logger)
        self.per_item_cost = per_item_cost
        self.default_cost = default_cost

    def change_order(
        self,       # type: OrderCostAdderBox
        order,      # type: ProductionOrder
    ):
        # type: (...) -> ProductionOrder

        order.cost = self._add_cost_to_items(order.items)
        return order

    def _add_cost_to_items(
        self,       # type: OrderCostAdderBox
        items,      # type: List[Item]
    ):
        # type: (...) -> int

        order_points = 0
        for item in items:
            if item.is_product():
                order_points += self._get_item_cost(item)
            else:
                order_points += self._get_item_cost(item) + self._add_cost_to_items(item.items)

        return order_points

    def _get_item_cost(
        self,       # type: OrderCostAdderBox
        item,       # type: Item
    ):
        # type: (...) -> int
        if item.part_code in self.per_item_cost:
            item.cost = self.per_item_cost[item.part_code] * item.multiplied_qty
        else:
            item.cost = self.default_cost * item.multiplied_qty

        return item.cost



from abc import ABCMeta, abstractmethod
from logging import Logger

from production.treebuilder import BoxConfiguration
from typing import Dict, Optional, Any

from production.box import ProductionBox


class BoxBuilder(object):
    __metaclass__ = ABCMeta

    def __init__(self, loggers):
        # type: (Dict[str, Logger]) -> None

        self.loggers = loggers

    @abstractmethod
    def build(self, config):
        # type: (BoxConfiguration) -> ProductionBox

        raise NotImplementedError()

    @staticmethod
    def get_extra(config, name, default):
        # type: (BoxConfiguration, str, Any) -> Any

        if name in config.extra_config:
            return config.extra_config[name]
        else:
            return default

    def get_logger(self, config):
        # type: (BoxConfiguration) -> Optional[Logger]

        if config.log_level in self.loggers:
            return self.loggers[config.log_level]
        else:
            return None



# -*- coding: utf-8 -*-

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/api/auth.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x15google/api/auth.proto\x12\ngoogle.api"l\n\x0e\x41uthentication\x12-\n\x05rules\x18\x03 \x03(\x0b\x32\x1e.google.api.AuthenticationRule\x12+\n\tproviders\x18\x04 \x03(\x0b\x32\x18.google.api.AuthProvider"\xa9\x01\n\x12\x41uthenticationRule\x12\x10\n\x08selector\x18\x01 \x01(\t\x12,\n\x05oauth\x18\x02 \x01(\x0b\x32\x1d.google.api.OAuthRequirements\x12 \n\x18\x61llow_without_credential\x18\x05 \x01(\x08\x12\x31\n\x0crequirements\x18\x07 \x03(\x0b\x32\x1b.google.api.AuthRequirement"^\n\x0bJwtLocation\x12\x10\n\x06header\x18\x01 \x01(\tH\x00\x12\x0f\n\x05query\x18\x02 \x01(\tH\x00\x12\x10\n\x06\x63ookie\x18\x04 \x01(\tH\x00\x12\x14\n\x0cvalue_prefix\x18\x03 \x01(\tB\x04\n\x02in"\x9a\x01\n\x0c\x41uthProvider\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x06issuer\x18\x02 \x01(\t\x12\x10\n\x08jwks_uri\x18\x03 \x01(\t\x12\x11\n\taudiences\x18\x04 \x01(\t\x12\x19\n\x11\x61uthorization_url\x18\x05 \x01(\t\x12.\n\rjwt_locations\x18\x06 \x03(\x0b\x32\x17.google.api.JwtLocation"-\n\x11OAuthRequirements\x12\x18\n\x10\x63\x61nonical_scopes\x18\x01 \x01(\t"9\n\x0f\x41uthRequirement\x12\x13\n\x0bprovider_id\x18\x01 \x01(\t\x12\x11\n\taudiences\x18\x02 \x01(\tBk\n\x0e\x63om.google.apiB\tAuthProtoP\x01ZEgoogle.golang.org/genproto/googleapis/api/serviceconfig;serviceconfig\xa2\x02\x04GAPIb\x06proto3'
)


_AUTHENTICATION = DESCRIPTOR.message_types_by_name["Authentication"]
_AUTHENTICATIONRULE = DESCRIPTOR.message_types_by_name["AuthenticationRule"]
_JWTLOCATION = DESCRIPTOR.message_types_by_name["JwtLocation"]
_AUTHPROVIDER = DESCRIPTOR.message_types_by_name["AuthProvider"]
_OAUTHREQUIREMENTS = DESCRIPTOR.message_types_by_name["OAuthRequirements"]
_AUTHREQUIREMENT = DESCRIPTOR.message_types_by_name["AuthRequirement"]
Authentication = _reflection.GeneratedProtocolMessageType(
    "Authentication",
    (_message.Message,),
    {
        "DESCRIPTOR": _AUTHENTICATION,
        "__module__": "google.api.auth_pb2"
        # @@protoc_insertion_point(class_scope:google.api.Authentication)
    },
)
_sym_db.RegisterMessage(Authentication)

AuthenticationRule = _reflection.GeneratedProtocolMessageType(
    "AuthenticationRule",
    (_message.Message,),
    {
        "DESCRIPTOR": _AUTHENTICATIONRULE,
        "__module__": "google.api.auth_pb2"
        # @@protoc_insertion_point(class_scope:google.api.AuthenticationRule)
    },
)
_sym_db.RegisterMessage(AuthenticationRule)

JwtLocation = _reflection.GeneratedProtocolMessageType(
    "JwtLocation",
    (_message.Message,),
    {
        "DESCRIPTOR": _JWTLOCATION,
        "__module__": "google.api.auth_pb2"
        # @@protoc_insertion_point(class_scope:google.api.JwtLocation)
    },
)
_sym_db.RegisterMessage(JwtLocation)

AuthProvider = _reflection.GeneratedProtocolMessageType(
    "AuthProvider",
    (_message.Message,),
    {
        "DESCRIPTOR": _AUTHPROVIDER,
        "__module__": "google.api.auth_pb2"
        # @@protoc_insertion_point(class_scope:google.api.AuthProvider)
    },
)
_sym_db.RegisterMessage(AuthProvider)

OAuthRequirements = _reflection.GeneratedProtocolMessageType(
    "OAuthRequirements",
    (_message.Message,),
    {
        "DESCRIPTOR": _OAUTHREQUIREMENTS,
        "__module__": "google.api.auth_pb2"
        # @@protoc_insertion_point(class_scope:google.api.OAuthRequirements)
    },
)
_sym_db.RegisterMessage(OAuthRequirements)

AuthRequirement = _reflection.GeneratedProtocolMessageType(
    "AuthRequirement",
    (_message.Message,),
    {
        "DESCRIPTOR": _AUTHREQUIREMENT,
        "__module__": "google.api.auth_pb2"
        # @@protoc_insertion_point(class_scope:google.api.AuthRequirement)
    },
)
_sym_db.RegisterMessage(AuthRequirement)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n\016com.google.apiB\tAuthProtoP\001ZEgoogle.golang.org/genproto/googleapis/api/serviceconfig;serviceconfig\242\002\004GAPI"
    _AUTHENTICATION._serialized_start = 37
    _AUTHENTICATION._serialized_end = 145
    _AUTHENTICATIONRULE._serialized_start = 148
    _AUTHENTICATIONRULE._serialized_end = 317
    _JWTLOCATION._serialized_start = 319
    _JWTLOCATION._serialized_end = 413
    _AUTHPROVIDER._serialized_start = 416
    _AUTHPROVIDER._serialized_end = 570
    _OAUTHREQUIREMENTS._serialized_start = 572
    _OAUTHREQUIREMENTS._serialized_end = 617
    _AUTHREQUIREMENT._serialized_start = 619
    _AUTHREQUIREMENT._serialized_end = 676
# @@protoc_insertion_point(module_scope)



class Part(object):
    def __init__(
        self,
        part_code,      # type: int
        product_name,   # type: str
        item_type,      # type: str
    ):
        # type: (...) -> None

        self.part_code = part_code
        self.product_name = product_name
        self.item_type = item_type



from threading import Lock
from typing import Optional, Any, List
from datetime import datetime

from application.model import UpdateType, UserUpdateData, PosUser
from application.model.configuration import Configurations
from application.model.customexception import (
    ErrorApplyingUpdate,
    DisabledFunctionality,
    SomeUserLogged,
    ErrorRetrievingUsers,
)
from application.repository.api import UserUpdaterAPIRepository
from application.repository.pos import UserUpdaterPOSRepository
from application.util import LoggerUtil, settings

logger = LoggerUtil.get_logger_name(__name__)


class UserUpdaterInteractor(object):
    def __init__(self, configs, working_on_update, api_repository, pos_repository, enabled):
        # type: (Configurations, Lock, Optional[UserUpdaterAPIRepository], UserUpdaterPOSRepository, bool) -> None

        self.logger = logger
        self.configs = configs
        self.update_type = UpdateType.user
        self.update_name = self.update_type.name
        self.local_configs = configs.updaters[self.update_name]
        self.api_repository = api_repository
        self.pos_repository = pos_repository
        self.working_on_update = working_on_update
        self.enabled = enabled

    def update_users(self, logged_users):
        # type: (List[int]) -> None

        settings.set_update_control(self.update_name, None)
        if not self.enabled:
            settings.set_update_control(self.update_name, "$UPDATE_DISABLED")
            raise DisabledFunctionality()
        try:
            self.logger.info("Waiting Updating Users")
            with self.working_on_update:
                self.logger.info("Updating Users")

                self._apply_update(logged_users)

                self.logger.info("Update fully concluded with success!")
        except (Exception,):
            self.logger.exception("Not expected exception on user update")
            raise

    def _get_updated_users(self):
        # type: () -> List[UserUpdateData]

        self.logger.info("Trying to obtain employees data...")
        updated_users = []

        received_employee_data = self.api_repository.get_employees_data()

        for user in received_employee_data:
            user_data = UserUpdateData(
                user_id=int(user.get("pos-user-id")),
                long_name=user.get("name").encode("utf-8"),
                password=user.get("pos-password"),
                nfc_id_code=user.get("nfc-id-code") or user.get("pos-password"),
                level=user.get("pos-access-level").get("code"),
                start_date=user.get("from-dt"),
                end_date=user.get("to-dt"),
                pay_rate=user.get("pay-rate-amount"),
            )
            user_data.fill_user_status()
            updated_users.append(user_data)

        self.logger.info("Package successfully obtained!")

        return updated_users

    def _apply_update(self, logged_users):
        # type: (Any) -> None

        try:
            self.logger.info("Applying update on database...")
            settings.set_update_control(self.update_name, "$GETTING_USERS_FROM_BACKOFFICE")
            settings.set_update_control("updating", True)

            updated_users = self._get_updated_users()
            settings.set_update_control(self.update_name, "$UPDATING_POS_USERS")
            pos_users = self.pos_repository.get_pos_users()
            users_to_disable = self._get_users_to_disable(updated_users, pos_users)
            if len(users_to_disable) > 0:
                users_disable_logged = self._get_users_disable_logged(users_to_disable, logged_users)
                if len(users_disable_logged) > 0:
                    raise SomeUserLogged(self._get_logged_user_names(pos_users, users_disable_logged))

                self.pos_repository.disable_users(users_to_disable)

            self.pos_repository.update_users(updated_users)

            last_user_update_id = self.pos_repository.get_last_user_update_id()
            self.pos_repository.insert_new_user_update(str(last_user_update_id + 1), "Users")

            settings.set_update_control(self.update_name, "$SUCCESS_UPDATE")
            self.logger.info("Package successfully applied!")
            self.pos_repository.mb_context.MB_EasyEvtSend("USERS_AND_LEVELS_MODIFIED", '', str(datetime.now()))
        except ErrorRetrievingUsers as ex:
            settings.set_update_control(self.update_name, ex.message)
            raise ErrorRetrievingUsers(ex)
        except SomeUserLogged as ex:
            self.logger.exception("Users {} are logged".format(ex.message))
            settings.set_update_control(self.update_name, "$SOME_USER_LOGGED")
            raise
        except Exception as ex:
            self.logger.exception("Error applying update")
            settings.set_update_control(self.update_name, "$UNKNOWN_ERROR")
            raise ErrorApplyingUpdate(ex)
        finally:
            settings.set_update_control("updating", False)

    def _get_users_to_disable(self, updated_users, pos_users):
        # type: (List[UserUpdateData], List[PosUser]) -> List[str]

        users_to_disable = []

        removed_users = self._get_removed_users(updated_users, pos_users)
        users_to_disable.extend(removed_users)

        expired_users = self._get_expired_users(updated_users, pos_users)
        users_to_disable.extend(expired_users)

        return users_to_disable

    @staticmethod
    def _get_removed_users(updated_users, pos_users):
        # type: (List[UserUpdateData], List[PosUser]) -> List[str]

        updated_users_id = [x.user_id for x in updated_users]
        removed_users = [str(x.user_id) for x in pos_users if int(x.level) <= 10 and x.user_id not in updated_users_id]
        return removed_users

    def _get_expired_users(self, updated_users, pos_users):
        # type: (List[UserUpdateData], List[PosUser]) -> List[str]

        today = datetime.today().replace(hour=0, minute=0, second=0, microsecond=0)
        users_with_end_date = [x for x in updated_users if x.end_date]
        expired_users = [x.user_id for x in users_with_end_date if self._parse_date(x.end_date) <= today]
        return [str(x.user_id) for x in pos_users if int(x.level) <= 10 and x.user_id in expired_users]

    @staticmethod
    def _parse_date(date):
        # type: (str) -> datetime

        return datetime.strptime(date, '%Y-%m-%d')

    @staticmethod
    def _get_logged_user_names(pos_users, logged_users):
        # type: (List[PosUser], List[int]) -> str

        user_names = [str(x.user_id) + " - " + x.long_name for x in pos_users if x.user_id in logged_users]
        return "\\".join(user_names)

    @staticmethod
    def _get_users_disable_logged(users_to_disable, logged_users):
        # type: (List[int], List[int]) -> List[int]

        return [int(x) for x in logged_users if str(x) in users_to_disable]



from abc import ABCMeta, abstractmethod

from typing import List  # noqa


class PrintReport(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def send_print(self, pos_id, preview, report_name, *report_params):
        # type: (str, bool, str, List[any]) -> str
        raise NotImplementedError()



from datetime import datetime

from typing import (
    Dict,
    List,
    Optional,
)
from order_api.dto import Order as OrderApi
from ._ServiceTender import ServiceTender


# noinspection PyShadowingBuiltins
class Table(object):

    def __init__(
        self,                       # type: Table
        table_id,                   # type: str
        order_id=None,              # type: Optional[str]
        user_id=None,               # type: Optional[str]
        table_mode=None,            # type: Optional[str]
        table_name=None,            # type: Optional[str]
        tabs_names=None,            # type: Optional[Dict[str, str]]
        orders=None,                # type: Optional[List[OrderApi]]
        total_amount=None,          # type: Optional[float]
        tip_rate=None,              # type: Optional[float]
        tip=None,                   # type: Optional[float]
        service_seats=None,         # type: Optional[int]
        total_per_seat=None,        # type: Optional[float]
        total_discounts=None,       # type: Optional[float]
        start_date=None,            # type: Optional[datetime]
        service_tenders=None,       # type: List[ServiceTender]
        status=0,                   # type: Optional[int]
        tip_base_amount=0,          # type: Optional[float]
    ):
        # type: (...) -> None

        self.table_id = table_id
        self.order_id = order_id
        self.user_id = user_id
        self.table_mode = table_mode
        self.table_name = table_name
        self.tabs_names = tabs_names or {}
        self.orders = orders or []
        self.total_amount = total_amount or 0.0
        self.tip_rate = tip_rate or 0.0
        self.tip = tip or 0.0
        self.service_seats = service_seats or 0
        self.total_per_seat = total_per_seat or 0.0
        self.total_discounts = total_discounts or 0.0
        self.start_date = start_date
        self.service_tenders = service_tenders or []
        self.status = status
        self.tip_base_amount = tip_base_amount

        self.sub_total_amount = 0.0         # type: float
        self.detailed_discounts = {}        # type: Dict
        self.detailed_payments = {}         # type: Dict
        self.user_name = None               # type: Optional[str]
        self.promotional_tenders = 0.0      # type: float
        self.detailed_tip_excluded = {}     # type: Dict
        self.discount_groups = {}           # type: Dict



from typing import (
    Dict,
    List,
    Optional,
    Union,
)

from nreports.repository import (
    I18NRepository,
    L10NRepository,
)
from nreports_api.shared import JSON
from ._Line import Line


class CommandLine(Line):

    def __init__(
        self,
        line: JSON,
    ) -> None:
        super().__init__(line)

        self.command = line.get("command")

    def inner_format(
        self,
        data: Dict[str, Union[JSON, Dict[str, List[Dict[str, JSON]]]]],
        i18n_repository: I18NRepository,
        l10n_repository: L10NRepository,
    ) -> List[Dict[str, str]]:

        return [{"command": self.command}]

    def get_inner_data_to_build(
        self,
    ) -> Dict[str, Optional[Dict[str, None]]]:

        return {}



import os
import json
import cfgtools
import sysactions
import logging

from cfgtools import Configuration
from typing import Union

logger = logging.getLogger("PosActions")
ask_document_pinpad = {}
request_void_reason_on = None
config = None
cancel_json = None


def get_config():
    # type: (...) -> Configuration

    global config

    if config is None:
        config = cfgtools.read(os.environ["LOADERCFG"])

    return config


def read_int_value_from_loader(
    loader,             # type: Configuration
    key,                # type: str
    default_value,      # type: int
):
    # type: (...) -> int

    try:
        value = int(loader.find_value(key))
        return value
    except (Exception,):
        return default_value


def get_ask_document_pinpad(
    pos_id,  # type: Union[int, str]
):
    # type: (...) -> bool

    global ask_document_pinpad

    pos_id = str(pos_id)
    if pos_id not in ask_document_pinpad:
        ask_document = sysactions.get_cfg(posid=pos_id).find_value("AskDocumentPinpad", "false").lower() == "true"
        ask_document_pinpad[pos_id] = ask_document

    return ask_document_pinpad[pos_id]


def must_require_void_reason(
    void_on,    # type: str
):
    # type: (...) -> bool

    global request_void_reason_on

    if request_void_reason_on is None:
        config_key = get_config().find_values("Customizations.requestVoidReasonOn")
        if config_key is not None:
            request_void_reason_on = config_key
        else:
            request_void_reason_on = ["VOID_CURRENT_ORDER"]

    return void_on in request_void_reason_on


def get_cancel_json():
    # type: (...) -> dict

    global cancel_json

    if cancel_json is None:
        json_path = os.path.join(
            os.environ.get('BUNDLEDIR'),
            "cancel.json",
        )
        if not os.path.exists(json_path):
            cancel_json = {}

        try:
            with open(json_path, "rb") as f:
                cancel_json = json.loads(f.read().decode("utf-8"))
        except Exception as e:
            logger.error(e)
            cancel_json = {}

    return cancel_json



import json

from pos_model import (
    SaleItem,
    TaxItem,
)
from typing import List

from common import (
    CalculateTaxesItemCommon,
    FiscalParameterController,
)


def build_item_taxes(
    sale_item,                      # type: SaleItem
    fiscal_parameter_controller,    # type: FiscalParameterController
    calculate_taxes_common,         # type: CalculateTaxesItemCommon
):
    # type: (...) -> List[TaxItem]

    taxes_calc = fiscal_parameter_controller.get_tip_params(part_code=sale_item.part_code)
    taxes = json.dumps(taxes_calc).encode("utf-8")

    base_reduction = fiscal_parameter_controller.get_tip_params(
        part_code=sale_item.part_code,
        taxes_name=['BASE_REDUCTION'],
    )
    base_reduction = None if len(base_reduction) == 0 else {sale_item.part_code: base_reduction[0]['rate']}

    cst_dict = _get_cst_dict(
        fiscal_parameter_controller=fiscal_parameter_controller,
        part_code=sale_item.part_code,
    )
    item_taxes = calculate_taxes_common.get_taxes_item(
        sale_line=sale_item,
        taxes_calc=taxes,
        base_reduction=base_reduction,
        cst=cst_dict
    )

    return item_taxes


def _get_cst_dict(
    fiscal_parameter_controller,
    part_code,
):
    cst = {}

    cst_pis = fiscal_parameter_controller.get_tax_value_from_loader(part_code, 'CST_PIS')
    cst['CST_PIS'] = cst_pis

    cst_icms = fiscal_parameter_controller.get_tax_value_from_loader(part_code, 'CST_ICMS')
    cst['CST_ICMS'] = cst_icms

    cst_cofins = fiscal_parameter_controller.get_tax_value_from_loader(part_code, 'CST_COFINS')
    cst['CST_COFINS'] = cst_cofins

    return cst



# cpf.py - functions for handling Costa Rica CPF numbers
# coding: utf-8
#
# Copyright (C) 2019 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""CPF (Cédula de Persona Física, Costa Rica physical person ID number).

The Cédula de Persona Física (CPF), also known as Cédula de Identidad is an
identifier of physical persons.

The number consists of 10 digits in the form 0P-TTTT-AAAA where P represents
the province, TTTT represents the volume (tomo) padded with zeroes on the
left, and AAAA represents the entry (asiento) also padded with zeroes on the
left.

It seems to be usual for the leading zeroes in each of the three parts to be
omitted.

More information:

* https://www.hacienda.go.cr/consultapagos/ayuda_cedulas.htm
* https://www.procomer.com/downloads/quiero/guia_solicitud_vuce.pdf (page 11)
* https://www.hacienda.go.cr/ATV/frmConsultaSituTributaria.aspx

>>> validate('3-0455-0175')
'0304550175'
>>> validate('30-1234-1234')
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('12345678')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('701610395')
'07-0161-0395'
>>> format('1-613-584')
'01-0613-0584'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace. Also adds padding zeroes if necessary.
    """
    number = clean(number, ' ').upper().strip()
    parts = number.split('-')
    if len(parts) == 3:
        # Pad each group with zeroes
        parts[0] = parts[0].zfill(2)
        parts[1] = parts[1].zfill(4)
        parts[2] = parts[2].zfill(4)
    number = ''.join(parts)
    if len(number) == 9:
        number = '0' + number  # Add leading zero
    return number


def validate(number):
    """Check if the number is a valid Costa Rica CPF number.

    This checks the length and formatting.
    """
    number = compact(number)
    if len(number) != 10:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if number[0] != '0':
        raise InvalidComponent()
    return number


def is_valid(number):
    """Check if the number is a valid Costa Rica CPF number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '-'.join([number[:2], number[2:6], number[6:]])



from ._PrinterViewBuilder import PrinterViewBuilder
from ._KdsViewBuilder import KdsViewBuilder
from ._PublishViewBuilder import PublishViewBuilder
from ._CommandViewBuilder import CommandViewBuilder
from ._CommonViewBuilder import get_extra, get_logger



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2003-2007, 2009-2011 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""A place to store TSIG keys."""

from dns._compat import maybe_decode, maybe_encode

import base64

import dns.name


def from_text(textring):
    """Convert a dictionary containing (textual DNS name, base64 secret) pairs
    into a binary keyring which has (dns.name.Name, binary secret) pairs.
    @rtype: dict"""

    keyring = {}
    for keytext in textring:
        keyname = dns.name.from_text(keytext)
        secret = base64.decodestring(maybe_encode(textring[keytext]))
        keyring[keyname] = secret
    return keyring


def to_text(keyring):
    """Convert a dictionary containing (dns.name.Name, binary secret) pairs
    into a text keyring which has (textual DNS name, base64 secret) pairs.
    @rtype: dict"""

    textring = {}
    for keyname in keyring:
        keytext = maybe_decode(keyname.to_text())
        secret = maybe_decode(base64.encodestring(keyring[keyname]))
        textring[keytext] = secret
    return textring



from typing import Optional


class AsyncTargetDto(object):
    def __init__(
        self,
        reply_subject: str,
        id_field: Optional[str],
    ):
        self.reply_subject = reply_subject
        self.id_field = id_field



import json
import logging

from application.repository.api import APIRepository
from dto import EndOfSaleNotesRequest
from nreports_api import NReports
from typing import (
    Dict,
    Optional,
)

logger = logging.getLogger(__name__)


class EndOfSaleNotesInteractor(object):

    def __init__(
        self,               # type: EndOfSaleNotesInteractor
        api_repository,     # type: APIRepository
        nreports_api,       # type: NReports
    ):
        # type: (...) -> None

        self.api_repository = api_repository
        self.nreports_api = nreports_api

    def execute(
        self,       # type: EndOfSaleNotesInteractor
        request,    # type: EndOfSaleNotesRequest
    ):
        # type: (...) -> Optional[Dict[str, str]]

        logger.info("Searching for end of sale notes. PosId/OrderId {}/{}".format(request.pos_id, request.order_id))
        response = self.api_repository.end_of_sale_notes(
            pos_id=request.pos_id,
            order_id=request.order_id,
        )

        logger.info("End of sale notes response: {}".format(response))
        if not response:
            return None

        parsed_data = self.nreports_api.request_report_json_parser(
            pos_id=None,
            report_json=json.dumps(response),
        )

        return parsed_data



from typing import Any


class OrderItem(object):
    def __init__(
        self,
        name,           # type: unicode
        quantity,       # type: float
        pcode,          # type: int
        unit_price,     # type: float
        context,        # type: str
    ):
        # type: (...) -> None

        self.name = name
        self.quantity = quantity
        self.pcode = pcode
        self.unit_price = unit_price
        self.context = context

    def __getitem__(
        self,
        key,    # type: str
    ):
        # type: (...) -> Any

        return self.__getattribute__(key)



from typing import Dict, Any
from pos_model import Order
from _NfceXmlPartBuilder import NfceXmlPartBuilder
from _ContextKeys import ContextKeys


class InfIntermedBuilder(NfceXmlPartBuilder):

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        delivery_partner = context[ContextKeys.delivery_partner]
        if delivery_partner is None:
            return ""

        partner_id = delivery_partner.get("partnerId")
        cnpj = delivery_partner.get("cnpj")
        xml = "<infIntermed><CNPJ>{0}</CNPJ><idCadIntTran>{1}</idCadIntTran></infIntermed>"
        return xml.format(cnpj, partner_id)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore

"""
OpenTelemetry Base Distribution (Distro)
"""

from abc import ABC, abstractmethod
from logging import getLogger

from pkg_resources import EntryPoint

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor

_LOG = getLogger(__name__)


class BaseDistro(ABC):
    """An ABC for distro"""

    _instance = None

    def __new__(cls, *args, **kwargs):

        if cls._instance is None:
            cls._instance = object.__new__(cls, *args, **kwargs)

        return cls._instance

    @abstractmethod
    def _configure(self, **kwargs):
        """Configure the distribution"""

    def configure(self, **kwargs):
        """Configure the distribution"""
        self._configure(**kwargs)

    def load_instrumentor(  # pylint: disable=no-self-use
        self, entry_point: EntryPoint, **kwargs
    ):
        """Takes a collection of instrumentation entry points
        and activates them by instantiating and calling instrument()
        on each one.

        Distros can override this method to customize the behavior by
        inspecting each entry point and configuring them in special ways,
        passing additional arguments, load a replacement/fork instead,
        skip loading entirely, etc.
        """
        instrumentor: BaseInstrumentor = entry_point.load()
        instrumentor().instrument(**kwargs)


class DefaultDistro(BaseDistro):
    def _configure(self, **kwargs):
        pass


__all__ = ["BaseDistro", "DefaultDistro"]
